"""liquidctl drivers for fifth generation Asetek 690LC liquid coolers.

Supported devices:

- EVGA CLC (120 CL12, 240, 280 or 360); modern generic Asetek 690LC
- NZXT Kraken X (X31, X41 or X61); legacy generic Asetek 690LC
- NZXT Kraken X (X40 or X60); legacy generic Asetek 690LC
- Corsair Hydro H80i GT, H100i GTX or H110i GTX
- Corsair Hydro H80i v2, H100i v2 or H115i

Copyright Jonas Malaco and contributors

Incorporates or uses as reference work by Kristóf Jakab, Sean Nelson
and Chris Griffith.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging

import usb

from liquidctl.driver.usb import UsbDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.keyval import RuntimeStorage
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

_CMD_RUNTIME = 0x10
_CMD_PROFILE = 0x11
_CMD_OVERRIDE = 0x12
_CMD_PUMP_PWM = 0x13
_CMD_LUID = 0x14
_CMD_READ_ONLY_RUNTIME = 0x20
_CMD_STORE_SETTINGS = 0x21
_CMD_EXTERNAL_TEMPERATURE = 0x22

_FIXED_SPEED_CHANNELS = {    # (message type, minimum duty, maximum duty)
    'pump':  (_CMD_PUMP_PWM, 50, 100),  # min/max must correspond to _MIN/MAX_PUMP_SPEED_CODE
}
_VARIABLE_SPEED_CHANNELS = {  # (message type, minimum duty, maximum duty)
    'fan':   (_CMD_PROFILE, 0, 100)
}
_MAX_PROFILE_POINTS = 6
_CRITICAL_TEMPERATURE = 60
_HIGH_TEMPERATURE = 45
_MIN_PUMP_SPEED_CODE = 0x32
_MAX_PUMP_SPEED_CODE = 0x42
_READ_ENDPOINT = 0x82
_READ_LENGTH = 32
_WRITE_ENDPOINT = 0x2

_LEGACY_FIXED_SPEED_CHANNELS = {    # (message type, minimum duty, maximum duty)
    'fan':  (_CMD_OVERRIDE, 0, 100),
    'pump':  (_CMD_PUMP_PWM, 50, 100),
}

# USBXpress specific control parameters; from the USBXpress SDK
# (Customization/CP21xx_Customization/AN721SW_Linux/silabs_usb.h)
_USBXPRESS_REQUEST = 0x02
_USBXPRESS_FLUSH_BUFFERS = 0x01
_USBXPRESS_CLEAR_TO_SEND = 0x02
_USBXPRESS_NOT_CLEAR_TO_SEND = 0x04
_USBXPRESS_GET_PART_NUM = 0x08

# Unknown control parameters; from Craig's libSiUSBXp and OpenCorsairLink
_UNKNOWN_OPEN_REQUEST = 0x00
_UNKNOWN_OPEN_VALUE = 0xffff

# Control request type
_USBXPRESS = usb.util.CTRL_OUT | usb.util.CTRL_TYPE_VENDOR | usb.util.CTRL_RECIPIENT_DEVICE


class _Base690Lc(UsbDriver):
    """Common methods for Asetek 690LC devices."""

    _LEGACY_690LC = False

    @classmethod
    def probe(cls, handle, legacy_690lc=False, **kwargs):
        """Probe `handle` and yield corresponding driver instances."""
        if legacy_690lc != cls._LEGACY_690LC:
            return
        yield from super().probe(handle, **kwargs)

    @classmethod
    def find_supported_devices(cls, **kwargs):
        """Find devices specifically compatible with this driver.

        Automatically sets the appropriate value for `legacy_690lc`.
        """

        return super().find_supported_devices(legacy_690lc=cls._LEGACY_690LC, **kwargs)

    def _configure_flow_control(self, clear_to_send):
        """Set the software clear-to-send flow control policy for device."""
        _LOGGER.debug('set clear to send = %s', clear_to_send)
        if clear_to_send:
            self.device.ctrl_transfer(_USBXPRESS, _USBXPRESS_REQUEST, _USBXPRESS_CLEAR_TO_SEND)
        else:
            self.device.ctrl_transfer(_USBXPRESS, _USBXPRESS_REQUEST, _USBXPRESS_NOT_CLEAR_TO_SEND)

    def _begin_transaction(self):
        """Begin a new transaction before writing to the device."""
        _LOGGER.debug('begin transaction')
        self.device.claim()
        self.device.ctrl_transfer(_USBXPRESS, _USBXPRESS_REQUEST, _USBXPRESS_FLUSH_BUFFERS)

    def _write(self, data):
        self.device.write(_WRITE_ENDPOINT, data)

    def _end_transaction_and_read(self):
        """End the transaction by reading from the device.

        According to the official documentation, as well as Craig's open-source
        implementation (libSiUSBXp), it should be necessary to check the queue
        size and read the data in chunks.  However, leviathan and its
        derivatives seem to work fine without this complexity; we also
        successfully follow this approach.
        """

        msg = self.device.read(_READ_ENDPOINT, _READ_LENGTH)
        self.device.release()
        return msg

    def _configure_device(self, color1=[0, 0, 0], color2=[0, 0, 0], color3=[255, 0, 0],
                          alert_temp=_HIGH_TEMPERATURE, interval1=0, interval2=0,
                          blackout=False, fading=False, blinking=False, enable_alert=True):
        self._write([0x10] + color1 + color2 + color3
                    + [alert_temp, interval1, interval2, not blackout, fading,
                       blinking, enable_alert, 0x00, 0x01])

    def _prepare_profile(self, profile, min_duty, max_duty, max_points):
        opt = list(profile)
        size = len(opt)
        if size < 1:
            raise ValueError('at least one PWM point required')
        elif size > _MAX_PROFILE_POINTS:
            raise ValueError(f'too many PWM points ({size}), only {max_points} supported')
        for i, (temp, duty) in enumerate(opt):
            opt[i] = (temp, clamp(duty, min_duty, max_duty))
        missing = max_points - size
        if missing:
            # Some issues were observed when padding with (0°C, 0%), though
            # they were hard to reproduce.  So far it *seems* that in some
            # instances the device will store the last "valid" profile index
            # somewhere, and would need another call to initialize() to clear
            # that up.  Padding with (CRIT, 100%) appears to avoid all issues,
            # at least within the reasonable range of operating temperatures.
            _LOGGER.info('filling missing %d PWM points with (60°C, 100%%)', missing)
            opt = opt + [(_CRITICAL_TEMPERATURE, 100)]*missing
        return opt

    def connect(self, **kwargs):
        """Connect to the device.

        Enables the device to send data to the host.
        """

        ret = super().connect(**kwargs)
        self._configure_flow_control(clear_to_send=True)
        return ret

    def initialize(self, **kwargs):
        """Initialize the device."""
        self._begin_transaction()
        self._configure_device()
        self._end_transaction_and_read()

    def disconnect(self, **kwargs):
        """Disconnect from the device.

        Implementation note: unlike SI_Close is supposed to do,¹ do not send
        _USBXPRESS_NOT_CLEAR_TO_SEND to the device.  This allows one program to
        disconnect without stopping reads from another.

        Surrounding device.read() with _USBXPRESS_[NOT_]CLEAR_TO_SEND would
        make more sense, but there seems to be a yet unknown minimum delay
        necessary for that to work reliably.

        ¹ https://github.com/craigshelley/SiUSBXp/blob/master/SiUSBXp.c
        """

        super().disconnect(**kwargs)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


class _ModernBase690Lc(_Base690Lc):

    def _persist(self):
        _LOGGER.debug('instructing device to persist current settings')
        self._begin_transaction()
        self._write([_CMD_STORE_SETTINGS])
        self._end_transaction_and_read()

    def initialize(self, non_volatile=False, **kwargs):
        super().initialize(**kwargs)
        if non_volatile:
            self._persist()

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        self._begin_transaction()
        self._write([_CMD_LUID, 0, 0, 0])
        msg = self._end_transaction_and_read()
        firmware = '{}.{}.{}.{}'.format(*tuple(msg[0x17:0x1b]))
        return [
            ('Liquid temperature', msg[10] + msg[14]/10, '°C'),
            ('Fan speed', msg[0] << 8 | msg[1], 'rpm'),
            ('Pump speed', msg[8] << 8 | msg[9], 'rpm'),
            ('Firmware version', firmware, '')
        ]

    def set_color(self, channel, mode, colors, time_per_color=1,
                  time_off=None, alert_threshold=_HIGH_TEMPERATURE, alert_color=[255, 0, 0],
                  speed=3, non_volatile=False, **kwargs):
        """Set the color mode for a specific channel."""
        # keyword arguments may have been forwarded from cli args and need parsing
        colors = list(colors)
        self._begin_transaction()
        if mode == 'rainbow':
            if isinstance(speed, str):
                speed = int(speed)
            self._write([0x23, clamp(speed, 1, 6)])
            # make sure to clear blinking or... chaos
            self._configure_device(alert_temp=clamp(alert_threshold, 0, 100), color3=alert_color)
        elif mode == 'fading':
            self._configure_device(fading=True, color1=colors[0], color2=colors[1],
                                   interval1=clamp(time_per_color, 1, 255),
                                   alert_temp=clamp(alert_threshold, 0, 100), color3=alert_color)
            self._write([0x23, 0])
        elif mode == 'blinking':
            if time_off is None:
                time_off = time_per_color
            self._configure_device(blinking=True, color1=colors[0],
                                   interval1=clamp(time_off, 1, 255),
                                   interval2=clamp(time_per_color, 1, 255),
                                   alert_temp=clamp(alert_threshold, 0, 100), color3=alert_color)
            self._write([0x23, 0])
        elif mode == 'fixed':
            self._configure_device(color1=colors[0], alert_temp=clamp(alert_threshold, 0, 100),
                                   color3=alert_color)
            self._write([0x23, 0])
        elif mode == 'blackout':  # stronger than just 'off', suppresses alerts and rainbow
            self._configure_device(blackout=True, alert_temp=clamp(alert_threshold, 0, 100),
                                   color3=alert_color)
        else:
            raise KeyError(f'unknown lighting mode {mode}')
        self._end_transaction_and_read()
        if non_volatile:
            self._persist()

    def set_speed_profile(self, channel, profile, non_volatile=False, **kwargs):
        """Set channel to follow a speed duty profile."""
        mtype, dmin, dmax = _VARIABLE_SPEED_CHANNELS[channel]
        adjusted = self._prepare_profile(profile, dmin, dmax, _MAX_PROFILE_POINTS)
        for temp, duty in adjusted:
            _LOGGER.info('setting %s PWM point: (%d°C, %d%%), device interpolated',
                         channel, temp, duty)
        temps, duties = map(list, zip(*adjusted))
        self._begin_transaction()
        self._write([mtype, 0] + temps + duties)
        self._end_transaction_and_read()
        if non_volatile:
            self._persist()

    def set_fixed_speed(self, channel, duty, non_volatile=False, **kwargs):
        """Set channel to a fixed speed duty."""
        if channel == 'fan':
            # While devices seem to recognize a specific channel for fixed fan
            # speeds (mtype == 0x12), its use can later conflict with custom
            # profiles.
            # Note for a future self: the conflict can be cleared with
            # *another* call to initialize(), i.e.  with another
            # configuration command.
            _LOGGER.info('using a flat profile to set %s to a fixed duty', channel)
            self.set_speed_profile(channel, [(0, duty), (_CRITICAL_TEMPERATURE - 1, duty)], non_volatile, **kwargs)
            return
        mtype, dmin, dmax = _FIXED_SPEED_CHANNELS[channel]
        duty = clamp(duty, dmin, dmax)
        total_levels = _MAX_PUMP_SPEED_CODE - _MIN_PUMP_SPEED_CODE + 1
        level = round((duty - dmin)/(dmax - dmin)*total_levels)
        effective_duty = round(dmin + level*(dmax - dmin)/total_levels)
        _LOGGER.info('setting %s PWM duty to %d%% (level %d)', channel, effective_duty, level)
        self._begin_transaction()
        self._write([mtype, _MIN_PUMP_SPEED_CODE + level])
        self._end_transaction_and_read()
        if non_volatile:
            self._persist()

class Modern690Lc(_ModernBase690Lc):
    """Modern fifth generation Asetek 690LC cooler."""

    _MATCHES = [
        (0x2433, 0xb200, 'Asetek 690LC (assuming EVGA CLC)', {}),
    ]

    def downgrade_to_legacy(self):
        """Take the device handle and return a new Legacy690Lc instance for it.

        This method returns a new instance that takes the device handle from
        `self`.  Because of this, the caller should immediately discard `self`,
        as it is no longer valid to call any of its methods or access any of
        its properties.

        While it is sometimes possible to downgrade a device that has seen
        modern traffic since it booted, this will generally not work.
        Additionally, no attempt to disconnect from the device is made while
        downgrading the instance.

        Thus, callers are strongly advised to only call this function before
        connecting to the device from this instance and, in fact, before
        calling any other methods at all on the device, from any instance.

        Finally, this method is not yet considered stable and its signature
        and/or behavior may change.  Callers should follow the development of
        liquidctl and the stabilization of this API.
        """
        legacy = Legacy690Lc(self.device, self._description)
        self.device = None
        self._description = None
        return legacy


class Legacy690Lc(_Base690Lc):
    """Legacy fifth generation Asetek 690LC cooler."""

    _MATCHES = [
        (0x2433, 0xb200, 'Asetek 690LC (assuming NZXT Kraken X)', {}),
    ]

    _LEGACY_690LC = True

    def __init__(self, device, description, **kwargs):
        super().__init__(device, description, **kwargs)
        # --device causes drivers to be instantiated even if they are later
        # discarded; defer instantiating the data storage until to connect()
        self._data = None

    def connect(self, runtime_storage=None, **kwargs):
        ret = super().connect(**kwargs)
        if runtime_storage:
            self._data = runtime_storage
        else:
            ids = f'vid{self.vendor_id:04x}_pid{self.product_id:04x}'
            loc = f'bus{self.bus}_port{"_".join(map(str, self.port))}'
            self._data = RuntimeStorage(key_prefixes=[ids, loc, 'legacy'])
        return ret

    def _set_all_fixed_speeds(self):
        self._begin_transaction()
        for channel in ['pump', 'fan']:
            mtype, dmin, dmax = _LEGACY_FIXED_SPEED_CHANNELS[channel]
            duty = clamp(self._data.load(f'{channel}_duty', of_type=int, default=dmax), dmin, dmax)
            _LOGGER.info('setting %s duty to %d%%', channel, duty)
            self._write([mtype, duty])
        return self._end_transaction_and_read()

    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        self._data.store('pump_duty', None)
        self._data.store('fan_duty', None)
        self._set_all_fixed_speeds()
        _warn_on_unsupported_option(**kwargs)

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        msg = self._set_all_fixed_speeds()
        firmware = '{}.{}.{}.{}'.format(*tuple(msg[0x17:0x1b]))
        return [
            ('Liquid temperature', msg[10] + msg[14]/10, '°C'),
            ('Fan speed', msg[0] << 8 | msg[1], 'rpm'),
            ('Pump speed', msg[8] << 8 | msg[9], 'rpm'),
            ('Firmware version', firmware, '')
        ]

    def set_color(self, channel, mode, colors, time_per_color=None, time_off=None,
                  alert_threshold=_HIGH_TEMPERATURE, alert_color=[255, 0, 0],
                  **kwargs):
        """Set the color mode for a specific channel."""
        # keyword arguments may have been forwarded from cli args and need parsing
        colors = list(colors)
        self._begin_transaction()
        if mode == 'fading':
            if time_per_color is None:
                time_per_color = 5
            self._configure_device(fading=True, color1=colors[0], color2=colors[1],
                                   interval1=clamp(time_per_color, 1, 255),
                                   alert_temp=clamp(alert_threshold, 0, 100), color3=alert_color)
        elif mode == 'blinking':
            if time_per_color is None:
                time_per_color = 1
            if time_off is None:
                time_off = time_per_color
            self._configure_device(blinking=True, color1=colors[0],
                                   interval1=clamp(time_off, 1, 255),
                                   interval2=clamp(time_per_color, 1, 255),
                                   alert_temp=clamp(alert_threshold, 0, 100), color3=alert_color)
        elif mode == 'fixed':
            self._configure_device(color1=colors[0], alert_temp=clamp(alert_threshold, 0, 100),
                                   color3=alert_color)
        elif mode == 'blackout':  # stronger than just 'off', suppresses alerts and rainbow
            self._configure_device(blackout=True, alert_temp=clamp(alert_threshold, 0, 100),
                                   color3=alert_color)
        else:
            raise KeyError(f'unsupported lighting mode {mode}')
        self._end_transaction_and_read()
        _warn_on_unsupported_option(**kwargs)

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""
        mtype, dmin, dmax = _LEGACY_FIXED_SPEED_CHANNELS[channel]
        duty = clamp(duty, dmin, dmax)
        self._data.store(f'{channel}_duty', duty)
        self._set_all_fixed_speeds()
        _warn_on_unsupported_option(**kwargs)

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


class Hydro690Lc(_ModernBase690Lc):
    """Corsair-branded fifth generation Asetek 690LC cooler."""

    _MATCHES = [
        (0x1b1c, 0x0c02, 'Corsair Hydro H80i GT (experimental)', {}),
        (0x1b1c, 0x0c03, 'Corsair Hydro H100i GTX (experimental)', {}),
        (0x1b1c, 0x0c07, 'Corsair Hydro H110i GTX (experimental)', {}),
        (0x1b1c, 0x0c08, 'Corsair Hydro H80i v2', {}),
        (0x1b1c, 0x0c09, 'Corsair Hydro H100i v2', {}),
        (0x1b1c, 0x0c0a, 'Corsair Hydro H115i', {}),
    ]

    def set_color(self, channel, mode, colors, **kwargs):
        """Set the color mode for a specific channel."""
        if mode == 'rainbow':
            raise KeyError(f'unsupported lighting mode {mode}')
        super().set_color(channel, mode, colors, **kwargs)


def _warn_on_unsupported_option(**kwargs):
    if kwargs.get('non_volatile'):
        _LOGGER.warning('device does not support non-volatile settings, skipping')


# deprecated aliases
AsetekDriver = Modern690Lc
LegacyAsetekDriver = Legacy690Lc
CorsairAsetekDriver = Hydro690Lc



# uses the psf/black style

import pytest
import os

from _testutils import MockHidapiDevice, MockPyusbDevice, Report

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.kraken3 import KrakenX3, KrakenZ3
from liquidctl.driver.kraken3 import (
    _COLOR_CHANNELS_KRAKENX,
    _SPEED_CHANNELS_KRAKENX,
    _COLOR_CHANNELS_KRAKENZ,
    _SPEED_CHANNELS_KRAKENZ,
    _HWMON_CTRL_MAPPING_KRAKENX,
    _HWMON_CTRL_MAPPING_KRAKENZ,
)
from test_krakenz3_response import krakenz3_response

from liquidctl.util import HUE2_MAX_ACCESSORIES_IN_CHANNEL as MAX_ACCESSORIES
from liquidctl.util import Hue2Accessory


# https://github.com/liquidctl/liquidctl/issues/160#issuecomment-664044103
X3_SAMPLE_STATUS = bytes.fromhex(
    "7502200036000B51535834353320012101A80635350000000000000000000000"
    "0000000000000000000000000000000000000000000000000000000000000000"
)
# https://github.com/liquidctl/liquidctl/issues/160#issue-665781804
X3_FAULTY_STATUS = bytes.fromhex(
    "7502200036000B5153583435332001FFFFCC0A64640000000000000000000000"
    "0000000000000000000000000000000000000000000000000000000000000000"
)

Z3_SAMPLE_STATUS = bytes.fromhex(
    "75012E0018001051393434363731011803690314140102000000000000000000"
    "0000000000000000000000000000000000000000000000000000000000000000"
)

test_curve_final_pwm = [
    30,
    32,
    34,
    36,
    38,
    40,
    42,
    44,
    46,
    48,
    50,
    58,
    65,
    72,
    80,
    82,
    83,
    85,
    87,
    88,
    90,
    91,
    92,
    93,
    94,
    95,
    96,
    97,
    98,
    99,
    100,
    100,
    100,
    100,
    100,
    100,
    100,
    100,
    100,
    100,
]


@pytest.fixture
def mock_krakenx3():
    raw = MockKraken(raw_led_channels=len(_COLOR_CHANNELS_KRAKENX) - 1)
    dev = KrakenX3(
        raw,
        "Mock Kraken X73",
        speed_channels=_SPEED_CHANNELS_KRAKENX,
        color_channels=_COLOR_CHANNELS_KRAKENX,
        hwmon_ctrl_mapping=_HWMON_CTRL_MAPPING_KRAKENX,
    )

    dev.connect()
    return dev


@pytest.fixture
def mock_krakenz3():
    raw = MockKraken(raw_led_channels=1)
    dev = MockKrakenZ3(
        raw,
        "Mock Kraken Z73",
        speed_channels=_SPEED_CHANNELS_KRAKENZ,
        color_channels=_COLOR_CHANNELS_KRAKENZ,
        hwmon_ctrl_mapping=_HWMON_CTRL_MAPPING_KRAKENZ,
    )

    dev.connect()
    return dev


class MockKraken(MockHidapiDevice):
    def __init__(self, raw_led_channels):
        super().__init__()
        self.raw_led_channels = raw_led_channels

    def write(self, data):
        reply = bytearray(64)
        if data[0:2] == [0x10, 0x01]:
            reply[0:2] = [0x11, 0x01]
        elif data[0:2] == [0x20, 0x03]:
            reply[0:2] = [0x21, 0x03]
            reply[14] = self.raw_led_channels
            if self.raw_led_channels > 1:
                reply[15 + 1 * MAX_ACCESSORIES] = Hue2Accessory.KRAKENX_GEN4_RING.value
                reply[15 + 2 * MAX_ACCESSORIES] = Hue2Accessory.KRAKENX_GEN4_LOGO.value
        elif data[0:2] == [0x30, 0x01]:
            reply[0:2] = [0x31, 0x01]
            reply[0x18] = 50  # lcd brightness
            reply[0x1A] = 0  # lcd orientation
        elif data[0:2] == [0x32, 0x1]:  # setup bucket
            reply[14] = 0x1
        elif data[0:2] == [0x32, 0x2]:  # delete bucker
            reply[0:2] = [0x33, 0x02]
            reply[14] = 0x1
        elif data[0:2] == [0x38, 0x1]:  # switch bucket
            reply[14] = 0x1

        self.preload_read(Report(0, reply))
        return super().write(data)


class MockKrakenZ3(KrakenZ3):
    def __init__(self, device, description, speed_channels, color_channels, **kwargs):
        KrakenX3.__init__(self, device, description, speed_channels, color_channels, **kwargs)

        self.bulk_device = MockPyusbDevice(0x1E71, 0x3008)
        self.bulk_device.close_winusb_device = self.bulk_device.release

        self.orientation = 0
        self.brightness = 50

        self.screen_mode = None

    def set_screen(self, channel, mode, value, **kwargs):
        self.screen_mode = mode
        self.hid_data_index = 0
        self.bulk_data_index = 0

        super().set_screen(channel, mode, value, **kwargs)

        assert self.hid_data_index == len(
            krakenz3_response[self.screen_mode + "_hid"]
        ), f"Incorrect number of hid messages sent for mode: {mode}"

        if mode == "static" or mode == "gif":
            assert (
                self.bulk_data_index == 801
                if mode == "static"
                else len(krakenz3_response[self.screen_mode + "_bulk"])
            ), f"Incorrect number of bulk messages sent for mode: {mode}"

    def _write(self, data):
        if self.screen_mode:
            assert (
                data == krakenz3_response[self.screen_mode + "_hid"][self.hid_data_index]
            ), f"HID write failed, wrong data for mode: {self.screen_mode}, data index: {self.hid_data_index}"
            self.hid_data_index += 1
        return super()._write(data)

    def _bulk_write(self, data):
        fixed_data_index = self.bulk_data_index
        if (
            self.screen_mode == "static" and self.bulk_data_index > 1
        ):  # the rest of the message should be identical to index 1
            fixed_data_index = 1

        assert (
            data == krakenz3_response[self.screen_mode + "_bulk"][fixed_data_index]
        ), f"Bulk write failed, wrong data for mode: {self.screen_mode}, data index: {self.bulk_data_index}"
        self.bulk_data_index += 1
        return super()._bulk_write(data)


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True), (True, False)])
def test_krakenx3_initializes(mock_krakenx3, has_hwmon, direct_access, tmp_path):
    if has_hwmon:
        mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)

    # TODO check the result
    _ = mock_krakenx3.initialize(direct_access=direct_access)

    writes = len(mock_krakenx3.device.sent)
    if not has_hwmon or direct_access:
        assert writes == 4
    else:
        assert writes == 2


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenx3_reads_status_directly(mock_krakenx3, has_hwmon, direct_access):
    if has_hwmon:
        mock_krakenx3._hwmon = HwmonDevice(None, None)

    mock_krakenx3.device.preload_read(Report(0, X3_SAMPLE_STATUS))

    temperature, pump_speed, pump_duty = mock_krakenx3.get_status(direct_access=direct_access)

    assert temperature == ("Liquid temperature", 33.1, "°C")
    assert pump_speed == ("Pump speed", 1704, "rpm")
    assert pump_duty == ("Pump duty", 53, "%")


def test_krakenx3_reads_status_from_hwmon(mock_krakenx3, tmp_path):
    mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("33100\n")
    (tmp_path / "fan1_input").write_text("1704\n")
    (tmp_path / "pwm1").write_text("135\n")

    temperature, pump_speed, pump_duty = mock_krakenx3.get_status()

    assert temperature == ("Liquid temperature", 33.1, "°C")
    assert pump_speed == ("Pump speed", 1704, "rpm")
    assert pump_duty == ("Pump duty", pytest.approx(53, rel=1.0 / 255), "%")


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenx3_set_fixed_speeds_directly(mock_krakenx3, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")

    mock_krakenx3.set_fixed_speed("pump", 84, direct_access=direct_access)

    pump_report = mock_krakenx3.device.sent[0]

    assert pump_report.number == 0x72
    assert pump_report.data[3:43] == [84 for i in range(0, 39)] + [100]

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_krakenx3_set_fixed_speeds_hwmon(mock_krakenx3, has_support, tmp_path):
    mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1").write_text("0\n")
        (tmp_path / "pwm1_enable").write_text("0\n")

    mock_krakenx3.set_fixed_speed("pump", 84)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "1"
        assert (tmp_path / "pwm1").read_text() == "214"
    else:
        # Assert fallback to direct access
        pump_report = mock_krakenx3.device.sent[0]

        assert pump_report.number == 0x72
        assert pump_report.data[3:43] == [84 for i in range(0, 39)] + [100]


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenx3_set_speed_profile_directly(mock_krakenx3, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        for i in range(1, 40 + 1):
            (tmp_path / f"temp1_auto_point{i}_pwm").write_text("0")

    curve_profile = zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100])

    mock_krakenx3.set_speed_profile("pump", curve_profile, direct_access=direct_access)

    pump_report = mock_krakenx3.device.sent[0]

    assert pump_report.number == 0x72
    assert pump_report.data[3:43] == test_curve_final_pwm

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        for i in range(1, 40):
            assert (tmp_path / f"temp1_auto_point{i}_pwm").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_krakenx3_set_speed_profile_hwmon(mock_krakenx3, has_support, tmp_path):
    mock_krakenx3._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1_enable").write_text("0\n")
        for i in range(1, 40 + 1):
            (tmp_path / f"temp1_auto_point{i}_pwm").write_text("0")

    curve_profile = zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100])

    mock_krakenx3.set_speed_profile("pump", curve_profile)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "2"
        for i in range(1, 40 + 1):
            assert int((tmp_path / f"temp1_auto_point{i}_pwm").read_text()) == (
                test_curve_final_pwm[i - 1] * 255 // 100
            )
    else:
        # Assert fallback to direct access
        pump_report = mock_krakenx3.device.sent[0]

        assert pump_report.number == 0x72
        assert pump_report.data[3:43] == test_curve_final_pwm


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenx3_warns_on_faulty_temperature(mock_krakenx3, has_hwmon, direct_access, caplog):
    if has_hwmon:
        mock_krakenx3._hwmon = HwmonDevice(None, None)

    mock_krakenx3.device.preload_read(Report(0, X3_FAULTY_STATUS))

    _ = mock_krakenx3.get_status(direct_access=direct_access)
    assert "unexpected temperature reading" in caplog.text


def test_krakenx3_not_totally_broken(mock_krakenx3):
    """Reasonable example calls to untested APIs do not raise exceptions."""
    mock_krakenx3.initialize()
    mock_krakenx3.set_color(channel="ring", mode="fixed", colors=iter([[3, 2, 1]]), speed="fastest")
    mock_krakenx3.set_speed_profile(channel="pump", profile=iter([(20, 20), (30, 50), (40, 100)]))
    mock_krakenx3.set_fixed_speed(channel="pump", duty=50)


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenz3_reads_status_directly(mock_krakenz3, has_hwmon, direct_access):
    if has_hwmon:
        mock_krakenz3._hwmon = HwmonDevice(None, None)

    mock_krakenz3.device.preload_read(Report(0, Z3_SAMPLE_STATUS))

    temperature, pump_speed, pump_duty, fan_speed, fan_duty = mock_krakenz3.get_status(
        direct_access=direct_access
    )

    assert temperature == ("Liquid temperature", 24.3, "°C")
    assert pump_speed == ("Pump speed", 873, "rpm")
    assert pump_duty == ("Pump duty", 20, "%")
    assert fan_speed == ("Fan speed", 0, "rpm")
    assert fan_duty == ("Fan duty", 0, "%")


def test_krakenz3_reads_status_from_hwmon(mock_krakenz3, tmp_path):
    mock_krakenz3._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("33100\n")
    (tmp_path / "fan1_input").write_text("1704\n")
    (tmp_path / "pwm1").write_text("135\n")
    (tmp_path / "fan2_input").write_text("1704\n")
    (tmp_path / "pwm2").write_text("135\n")

    temperature, pump_speed, pump_duty, fan_speed, fan_duty = mock_krakenz3.get_status()

    assert temperature == ("Liquid temperature", 33.1, "°C")
    assert pump_speed == ("Pump speed", 1704, "rpm")
    assert pump_duty == ("Pump duty", pytest.approx(53, rel=1.0 / 255), "%")
    assert fan_speed == ("Fan speed", 1704, "rpm")
    assert fan_duty == ("Fan duty", pytest.approx(53, rel=1.0 / 255), "%")


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenz3_set_fixed_speeds_directly(mock_krakenz3, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mock_krakenz3._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        (tmp_path / "pwm2").write_text("0")
        (tmp_path / "pwm2_enable").write_text("0")

    mock_krakenz3.set_fixed_speed("pump", 84, direct_access=direct_access)
    mock_krakenz3.set_fixed_speed("fan", 50, direct_access=direct_access)

    pump_report, fan_report = mock_krakenz3.device.sent

    assert pump_report.number == 0x72
    assert pump_report.data[3:43] == [84 for i in range(0, 39)] + [100]
    assert fan_report.number == 0x72
    assert fan_report.data[3:43] == [50 for i in range(0, 39)] + [100]

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        assert (tmp_path / "pwm2_enable").read_text() == "0"
        assert (tmp_path / "pwm2").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_krakenz3_set_fixed_speeds_hwmon(mock_krakenz3, has_support, tmp_path):
    mock_krakenz3._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1").write_text("0\n")
        (tmp_path / "pwm1_enable").write_text("0\n")
        (tmp_path / "pwm2").write_text("0\n")
        (tmp_path / "pwm2_enable").write_text("0\n")

    mock_krakenz3.set_fixed_speed("pump", 84)
    mock_krakenz3.set_fixed_speed("fan", 50)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "1"
        assert (tmp_path / "pwm1").read_text() == "214"
        assert (tmp_path / "pwm2_enable").read_text() == "1"
        assert (tmp_path / "pwm2").read_text() == "127"
    else:
        # Assert fallback to direct access
        pump_report, fan_report = mock_krakenz3.device.sent

        assert pump_report.number == 0x72
        assert pump_report.data[3:43] == [84 for i in range(0, 39)] + [100]
        assert fan_report.number == 0x72
        assert fan_report.data[3:43] == [50 for i in range(0, 39)] + [100]


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_krakenz3_set_speed_profile_directly(mock_krakenz3, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mock_krakenz3._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        (tmp_path / "pwm2").write_text("0")
        (tmp_path / "pwm2_enable").write_text("0")
        for i in range(1, 40 + 1):
            (tmp_path / f"temp1_auto_point{i}_pwm").write_text("0")
            (tmp_path / f"temp2_auto_point{i}_pwm").write_text("0")

    mock_krakenz3.set_speed_profile(
        "pump", zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100]), direct_access=direct_access
    )
    mock_krakenz3.set_speed_profile(
        "fan", zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100]), direct_access=direct_access
    )

    pump_report, fan_report = mock_krakenz3.device.sent

    assert pump_report.number == 0x72
    assert pump_report.data[3:43] == test_curve_final_pwm
    assert fan_report.number == 0x72
    assert fan_report.data[3:43] == test_curve_final_pwm

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        assert (tmp_path / "pwm2_enable").read_text() == "0"
        assert (tmp_path / "pwm2").read_text() == "0"
        for i in range(1, 40):
            assert (tmp_path / f"temp1_auto_point{i}_pwm").read_text() == "0"
            assert (tmp_path / f"temp2_auto_point{i}_pwm").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_krakenz3_set_speed_profile_hwmon(mock_krakenz3, has_support, tmp_path):
    mock_krakenz3._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1_enable").write_text("0\n")
        (tmp_path / "pwm2_enable").write_text("0\n")
        for i in range(1, 40 + 1):
            (tmp_path / f"temp1_auto_point{i}_pwm").write_text("0")
            (tmp_path / f"temp2_auto_point{i}_pwm").write_text("0")

    mock_krakenz3.set_speed_profile("pump", zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100]))
    mock_krakenz3.set_speed_profile("fan", zip([20, 30, 34, 40, 50], [30, 50, 80, 90, 100]))

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "2"
        for i in range(1, 40 + 1):
            assert int((tmp_path / f"temp1_auto_point{i}_pwm").read_text()) == (
                test_curve_final_pwm[i - 1] * 255 // 100
            )
            assert int((tmp_path / f"temp2_auto_point{i}_pwm").read_text()) == (
                test_curve_final_pwm[i - 1] * 255 // 100
            )
    else:
        # Assert fallback to direct access
        pump_report, fan_report = mock_krakenz3.device.sent

        assert pump_report.number == 0x72
        assert pump_report.data[3:43] == test_curve_final_pwm
        assert fan_report.number == 0x72
        assert fan_report.data[3:43] == test_curve_final_pwm


def test_krakenz3_not_totally_broken(mock_krakenz3):
    """Reasonable example calls to untested APIs do not raise exceptions."""
    mock_krakenz3.initialize()
    mock_krakenz3.device.preload_read(Report(0, Z3_SAMPLE_STATUS))
    _ = mock_krakenz3.get_status()
    mock_krakenz3.set_speed_profile(channel="fan", profile=iter([(20, 20), (30, 50), (40, 100)]))
    mock_krakenz3.set_fixed_speed(channel="pump", duty=50)

    # set_screen should be the last set of functions called
    mock_krakenz3.set_screen("lcd", "liquid", None)
    mock_krakenz3.set_screen("lcd", "brightness", "60")
    mock_krakenz3.set_screen("lcd", "orientation", "90")
    mock_krakenz3.set_screen(
        "lcd", "static", os.path.join(os.path.dirname(os.path.abspath(__file__)), "yellow.jpg")
    )
    mock_krakenz3.set_screen(
        "lcd", "gif", os.path.join(os.path.dirname(os.path.abspath(__file__)), "rgb.gif")
    )



"""Monitor and control liquid coolers and other devices.

liquidctl provides facilities for monitoring and controlling liquid coolers and
other hardware monitoring or LED controller devices in Python:

    from liquidctl import find_liquidctl_devices

    # Find all connected and supported devices.
    devices = find_liquidctl_devices()

    for dev in devices:

        # Connect to the device. In this example we use a context manager, but
        # the connection can also be manually managed. The context manager
        # automatically calls `disconnect`; when managing the connection
        # manually, `disconnect` must eventually be called, even if an
        # exception is raised.
        with dev.connect():
            print(f'{dev.description} at {dev.bus}:{dev.address}:')

            # Devices should be initialized after every boot. In this example
            # we assume that this has not been done before.
            print('- initialize')
            init_status = dev.initialize()

            # Print all data returned by `initialize`.
            if init_status:
                for key, value, unit in init_status:
                    print(f'- {key}: {value} {unit}')

            # Get regular status information from the device.
            status = dev.get_status()

            # Print all data returned by `get_status`.
            print('- get status')
            for key, value, unit in status:
                print(f'- {key}: {value} {unit}')

            # For a particular device, set the pump LEDs to red.
            if 'Kraken' in dev.description:
                print('- set pump to radical red')
                radical_red = [0xff, 0x35, 0x5e]
                dev.set_color(channel='pump', mode='fixed', colors=[radical_red])

A command-line interface is also available:

    $ python -m liquidctl --help

Once the liquidctl package is installed, a `liquidctl` executable should also
be available:

    $ liquidctl --help

Copyright 2018–2023 Jonas Malaco, Marshall Asch, CaseySJ, Tom Frey, Andrew
Robertson, ParkerMc, Aleksa Savic, Shady Nawara and contributors

Some modules also incorporate or use as reference work by leaty, Ksenija
Stanojevic, Alexander Tong, Jens Neumaier, Kristóf Jakab, Sean Nelson, Chris
Griffith, notaz, realies and Thomas Pircher. This is mentioned in the module
docstring, along with appropriate additional copyright notices.

SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style

from liquidctl.driver import find_liquidctl_devices
from liquidctl.error import *
from liquidctl.version import __version__



"""liquidctl drivers for third generation NZXT Kraken X and M liquid coolers.

Kraken X (X42, X52, X62 and X72)
--------------------------------

These coolers house 5-th generation Asetek pumps with additional PCBs for
advanced control and RGB capabilites.

Kraken M22
----------

The Kraken M22 shares similar RGB funcionality to the X models of the same
generation, but has no liquid temperature sensor and no hability to report or
set fan or pump speeds.

Copyright Jonas Malaco and contributors

Incorporates work by leaty, Ksenija Stanojevic, Alexander Tong and Jens
Neumaier.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import itertools
import logging

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import clamp, normalize_profile, interpolate_profile, \
                           map_direction

_LOGGER = logging.getLogger(__name__)

_SPEED_CHANNELS = {  # (base, minimum duty, maximum duty)
    'fan':   (0x80, 0, 100),
    'pump':  (0xc0, 0, 100),
}

_STATUS_TEMPERATURE = 'Liquid temperature'
_STATUS_FAN_SPEED = 'Fan speed'
_STATUS_PUMP_SPEED = 'Pump speed'
_STATUS_FWVERSION = 'Firmware version'

# more aggressive than observed 4.0.3 and 6.2 firmware defaults
_RESET_FAN_PROFILE = [(20, 25), (30, 50), (50, 90), (60, 100)]
_RESET_PUMP_PROFILE = [(20, 50), (30, 60), (40, 90), (50, 100)]

_CRITICAL_TEMPERATURE = 60

_COLOR_CHANNELS = {
    'sync':     0x0,
    'logo':     0x1,
    'ring':     0x2,
}

_COLOR_MODES = {
    # (byte3/mode, byte2/reverse, byte4/modifier, min colors, max colors, only ring)
    'off':                           (0x00, 0x00, 0x00, 0, 0, False),
    'fixed':                         (0x00, 0x00, 0x00, 1, 1, False),
    'super-fixed':                   (0x00, 0x00, 0x00, 1, 9, False),  # independent logo + ring leds
    'fading':                        (0x01, 0x00, 0x00, 2, 8, False),
    'spectrum-wave':                 (0x02, 0x00, 0x00, 0, 0, False),
    'marquee-3':                     (0x03, 0x00, 0x00, 1, 1, True),
    'marquee-4':                     (0x03, 0x00, 0x08, 1, 1, True),
    'marquee-5':                     (0x03, 0x00, 0x10, 1, 1, True),
    'marquee-6':                     (0x03, 0x00, 0x18, 1, 1, True),
    'covering-marquee':              (0x04, 0x00, 0x00, 1, 8, True),
    'alternating':                   (0x05, 0x00, 0x00, 2, 2, True),
    'moving-alternating':            (0x05, 0x08, 0x00, 2, 2, True),
    'breathing':                     (0x06, 0x00, 0x00, 1, 8, False),  # colors for each step
    'super-breathing':               (0x06, 0x00, 0x00, 1, 9, False),  # one step, independent logo + ring leds
    'pulse':                         (0x07, 0x00, 0x00, 1, 8, False),
    'tai-chi':                       (0x08, 0x00, 0x00, 2, 2, True),
    'water-cooler':                  (0x09, 0x00, 0x00, 0, 0, True),
    'loading':                       (0x0a, 0x00, 0x00, 1, 1, True),
    'wings':                         (0x0c, 0x00, 0x00, 1, 1, True),
    'super-wave':                    (0x0d, 0x00, 0x00, 1, 8, True),  # independent ring leds
}

_ANIMATION_SPEEDS = {
    'slowest':  0x0,
    'slower':   0x1,
    'normal':   0x2,
    'faster':   0x3,
    'fastest':  0x4,
}

_READ_ENDPOINT = 0x81
_READ_LENGTH = 64
_WRITE_ENDPOINT = 0x1
_WRITE_LENGTH = 65


class Kraken2(UsbHidDriver):
    """Third generation NZXT Kraken X or M liquid cooler."""

    # support for hwmon: nzxt-kraken2, Linux 5.13

    DEVICE_KRAKENX = 'Kraken X'
    DEVICE_KRAKENM = 'Kraken M'
    _MATCHES = [
        (0x1e71, 0x170e, 'NZXT Kraken X (X42, X52, X62 or X72)', {
            'device_type': DEVICE_KRAKENX
        }),
        (0x1e71, 0x1715, 'NZXT Kraken M22', {
            'device_type': DEVICE_KRAKENM
        }),
    ]

    def __init__(self, device, description, device_type=DEVICE_KRAKENX, **kwargs):
        super().__init__(device, description)
        self.device_type = device_type
        self.supports_lighting = True
        self.supports_cooling = self.device_type != self.DEVICE_KRAKENM
        self._firmware_version = None  # read once necessary

    def initialize(self, **kwargs):
        """Initialize the device and the driver.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        if self.supports_cooling_profiles:
            # due to a firmware limitation the same set of temperatures must be
            # used on both channels; ensure that is always true, even if the
            # user later only changes one of them, by resetting the profiles
            self.set_speed_profile('fan', _RESET_FAN_PROFILE)
            self.set_speed_profile('pump', _RESET_PUMP_PROFILE)

        fw = self.firmware_version
        _LOGGER.debug('raw firmware version: %r', fw)

        # after the transition to the unified 5.x/6.x firmware, NZXT/CAM
        # simplified how they report the firmware version from its raw
        # 4-component form
        fw_human = f'{fw[0]}.{fw[3]}' if fw[0] >= 5 else f'{fw[0]}.{fw[2]}.{fw[3]}'
        return [('Firmware version', fw_human, '')]

    def _get_status_directly(self):
        msg = self._read()

        return [
            (_STATUS_TEMPERATURE, msg[1] + msg[2]/10, '°C'),
            (_STATUS_FAN_SPEED, msg[3] << 8 | msg[4], 'rpm'),
            (_STATUS_PUMP_SPEED, msg[5] << 8 | msg[6], 'rpm'),
        ]

    def _get_status_from_hwmon(self):
        return [
            (_STATUS_TEMPERATURE, self._hwmon.read_int('temp1_input') * 1e-3, '°C'),
            (_STATUS_FAN_SPEED, self._hwmon.read_int('fan1_input'), 'rpm'),
            (_STATUS_PUMP_SPEED, self._hwmon.read_int('fan2_input'), 'rpm'),
        ]

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        if self.device_type == self.DEVICE_KRAKENM:
            return []

        if self._hwmon and not direct_access:
            _LOGGER.info('bound to %s kernel driver, reading status from hwmon', self._hwmon.driver)
            return self._get_status_from_hwmon()

        if self._hwmon:
            _LOGGER.warning('directly reading the status despite %s kernel driver',
                            self._hwmon.driver)

        return self._get_status_directly()

    def set_color(self, channel, mode, colors, speed='normal', direction='forward', **kwargs):
        """Set the color mode for a specific channel."""

        if not self.supports_lighting:
            raise NotSupportedByDevice()

        if mode == 'super':
            _LOGGER.warning('deprecated mode, move to super-fixed, super-breathing or super-wave')
            mode = 'super-fixed'
        if 'backwards' in mode:
            _LOGGER.warning('deprecated mode, move to direction=backward option')
            mode = mode.replace('backwards-', '')
            direction = 'backward'

        mval, mod2, mod4, mincolors, maxcolors, ringonly = _COLOR_MODES[mode]
        mod2 += map_direction(direction, 0, 0x10)

        if ringonly and channel != 'ring':
            _LOGGER.warning('mode=%s unsupported with channel=%s, dropping to ring',
                            mode, channel)
            channel = 'ring'

        steps = self._generate_steps(colors, mincolors, maxcolors, mode, ringonly)
        sval = _ANIMATION_SPEEDS[speed]
        byte2 = mod2 | _COLOR_CHANNELS[channel]
        for i, leds in enumerate(steps):
            seq = i << 5
            byte4 = sval | seq | mod4
            logo = [leds[0][1], leds[0][0], leds[0][2]]
            ring = list(itertools.chain(*leds[1:]))
            self._write([0x2, 0x4c, byte2, mval, byte4] + logo + ring)

    def _generate_steps(self, colors, mincolors, maxcolors, mode, ringonly):
        colors = list(colors)
        if len(colors) < mincolors:
            raise ValueError(f'not enough colors for mode={mode}, at least {mincolors} required')
        elif maxcolors == 0:
            if len(colors) > 0:
                _LOGGER.warning('too many colors for mode=%s, none needed', mode)
            colors = [(0, 0, 0)]  # discard the input but ensure at least one step
        elif len(colors) > maxcolors:
            _LOGGER.warning('too many colors for mode=%s, dropping to %d',
                            mode, maxcolors)
            colors = colors[:maxcolors]
        # generate steps from mode and colors: usually each color set by the user generates
        # one step, where it is specified to all leds and the device handles the animation;
        # but in super mode there is a single step and each color directly controls a led
        if 'super' not in mode:
            steps = [(color,)*9 for color in colors]
        elif ringonly:
            steps = [[(0, 0, 0)] + colors]
        else:
            steps = [colors]
        return steps

    def set_speed_profile(self, channel, profile, **kwargs):
        """Set channel to follow a speed duty profile."""

        if not self.supports_cooling_profiles:
            raise NotSupportedByDevice()
        norm = normalize_profile(profile, _CRITICAL_TEMPERATURE)
        # due to a firmware limitation the same set of temperatures must be
        # used on both channels; we reduce the number of writes by trimming the
        # interval and/or resolution to the most useful range
        stdtemps = list(range(20, 50)) + list(range(50, 60, 2)) + [60]
        interp = [(t, interpolate_profile(norm, t)) for t in stdtemps]
        cbase, dmin, dmax = _SPEED_CHANNELS[channel]
        for i, (temp, duty) in enumerate(interp):
            duty = clamp(duty, dmin, dmax)
            _LOGGER.info('setting %s PWM duty to %d%% for liquid temperature >= %d°C',
                         channel, duty, temp)
            self._write([0x2, 0x4d, cbase + i, temp, duty])

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""

        if not self.supports_cooling:
            raise NotSupportedByDevice()
        elif self.supports_cooling_profiles:
            self.set_speed_profile(channel, [(0, duty), (59, duty), (60, 100), (100, 100)])
        else:
            self.set_instantaneous_speed(channel, duty)

    def set_instantaneous_speed(self, channel, duty, **kwargs):
        """Set channel to speed duty, but do not guarantee persistence."""

        if not self.supports_cooling:
            raise NotSupportedByDevice()
        cbase, dmin, dmax = _SPEED_CHANNELS[channel]
        duty = clamp(duty, dmin, dmax)
        _LOGGER.info('setting %s PWM duty to %d%%', channel, duty)
        self._write([0x2, 0x4d, cbase & 0x70, 0, duty])

    @property
    def supports_cooling_profiles(self):
        return self.supports_cooling and self.firmware_version[0] >= 3

    @property
    def firmware_version(self):
        if self._firmware_version is None:
            _ = self._read(clear_first=False)
        return self._firmware_version

    def _read(self, clear_first=True):
        if clear_first:
            self.device.clear_enqueued_reports()
        msg = self.device.read(_READ_LENGTH)
        self._firmware_version = tuple(msg[0xb:0xf])
        return msg

    def _write(self, data):
        padding = [0x0]*(_WRITE_LENGTH - len(data))
        self.device.write(data + padding)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


# deprecated aliases
KrakenTwoDriver = Kraken2



"""Access to Linux HWMON data.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style

import logging
import sys
from pathlib import Path

_LOGGER = logging.getLogger(__name__)
_IS_LINUX = sys.platform == "linux"


class HwmonDevice:
    """Unstable API."""

    __slots__ = [
        "driver",
        "path",
    ]

    def __init__(self, driver, path):
        self.driver = driver
        self.path = path

    @property
    def name(self):
        return self.path.name

    def has_attribute(self, name):
        return (self.path / name).is_file()

    def get_string(self, name):
        value = (self.path / name).read_text().rstrip()
        _LOGGER.debug("read %s: %s", name, value)
        return value

    def read_int(self, name):
        return int(self.get_string(name))

    def write_int(self, name, value):
        (self.path / name).write_text(str(value))

    @classmethod
    def from_hidraw(cls, path):
        """Find the `HwmonDevice` for `path`."""

        if not _IS_LINUX:
            return None

        if not path.startswith(b"/dev/hidraw"):
            _LOGGER.debug("cannot search hwmon device for %s: unsupported path", path)
            return None

        path = path.decode()
        name = path[5:]
        class_path = Path("/sys/class/hidraw", name)

        sys_device = class_path / "device"
        hwmon_devices = sys_device / "hwmon"

        if not hwmon_devices.exists():
            return None

        hwmon_paths = hwmon_devices.iterdir()
        path = next(hwmon_paths)

        if next(hwmon_paths, None):
            _LOGGER.debug("cannot pick hwmon device for %s: more than one alternative", path)
            return None

        # use resolve() to be compatible with Python < 3.9
        driver = (sys_device / "driver").resolve().name

        return HwmonDevice(driver, path)



import pytest
from _testutils import VirtualBusDevice, VirtualControlMode

import json
import sys

import liquidctl.cli


@pytest.fixture
def main(monkeypatch, capsys):
    """Return a function f(*args) to run main with `args` as `sys.argv`."""
    def call_with_args(*args):
        monkeypatch.setattr(sys, 'argv', args)
        try:
            liquidctl.cli.main()
        except SystemExit as exit:
            code = exit.code
        else:
            code = 0
        out, err = capsys.readouterr()
        return code, out, err
    return call_with_args


def test_json_list(main):
    code, out, _ = main('test', '--bus', 'virtual', 'list', '--json')
    assert code == 0

    got = json.loads(out)
    exp = [
        {
            'description': 'Virtual Bus Device',
            'vendor_id': 0x1234,
            'product_id': 0xabcd,
            'release_number': None,
            'serial_number': None,
            'bus': 'virtual',
            'address': 'virtual_address',
            'port': None,
            'driver': 'VirtualBusDevice',
            'experimental': True,
        }
    ]
    assert got == exp


def test_json_initialize(main):
    code, out, _ = main('test', '--bus', 'virtual', 'initialize', '--json')
    assert code == 0

    got = json.loads(out)
    exp = [
        {
            'bus': 'virtual',
            'address': 'virtual_address',
            'description': 'Virtual Bus Device',
            'status': [
                { 'key': 'Firmware version', 'value': '3.14.16', 'unit': '' },
            ]
        }
    ]
    assert got == exp


def test_json_status(main):
    code, out, _ = main('test', '--bus', 'virtual', 'status', '--json')
    assert code == 0

    got = json.loads(out)
    exp = [
        {
            'bus': 'virtual',
            'address': 'virtual_address',
            'description': 'Virtual Bus Device',
            'status': [
                { 'key': 'Temperature', 'value': 30.4, 'unit': '°C' },
                { 'key': 'Fan control mode', 'value': 'VirtualControlMode.QUIET', 'unit': '' },
                { 'key': 'Animation', 'value': None, 'unit': '' },
                { 'key': 'Uptime', 'value': 66192.0, 'unit': 's' },
                { 'key': 'Hardware mode', 'value': True, 'unit': '' },
            ]
        }
    ]
    assert got == exp



import pytest

from pathlib import Path

from liquidctl.driver.smbus import LinuxI2c, LinuxI2cBus, SmbusDriver


class Canary(SmbusDriver):
    """Canary driver to reveal if SMBus probing is taking place."""

    @classmethod
    def probe(cls, smbus, **kwargs):
        yield Canary(smbus, 'Canary', vendor_id=-1, product_id=-1, address=-1)

    def connect(self, **kwargs):
        # there is no justification for calling connect on this test driver
        raise RuntimeError('forbidden')

    def __repr__(self):
        return repr(self._smbus)


def replace_smbus(replacement, monkeypatch):
    import liquidctl.driver.smbus

    # fragile hack: since messing with sys.meta_path (PEP 302) is tricky in a
    # pytest context, get into liquidctl.driver.smbus module and replace the
    # imported SMBus class after the fact
    monkeypatch.setattr(liquidctl.driver.smbus, 'SMBus', replacement)

    return replacement


@pytest.fixture
def emulated_smbus(monkeypatch):
    """Replace the SMBus implementation in liquidctl.driver.smbus."""

    class SMBus:
        def __init__(self, number):
            pass

    return replace_smbus(SMBus, monkeypatch)


def test__helper_fixture_replaces_real_smbus_implementation(emulated_smbus, tmpdir):
    i2c_dev = Path(tmpdir.mkdir('i2c-9999'))  # unlikely to be valid
    bus = LinuxI2cBus(i2c_dev=i2c_dev)

    bus.open()

    assert type(bus._smbus) == emulated_smbus


def test_filter_by_usb_port_yields_no_devices(emulated_smbus):
    discovered = Canary.find_supported_devices(usb_port='usb1')
    assert discovered == []


def test_aborts_if_sysfs_is_missing_devices(emulated_smbus, tmpdir):
    empty = tmpdir.mkdir('sys').mkdir('bus').mkdir('i2c')
    virtual_bus = LinuxI2c(i2c_root=empty)

    discovered = Canary.find_supported_devices(root_bus=virtual_bus)
    assert discovered == []


def test_finds_a_device(emulated_smbus, tmpdir):
    i2c_root = tmpdir.mkdir('sys').mkdir('bus').mkdir('i2c')
    i2c_root.mkdir('devices').mkdir('i2c-42')

    virtual_bus = LinuxI2c(i2c_root=i2c_root)

    discovered = Canary.find_supported_devices(root_bus=virtual_bus)
    assert len(discovered) == 1
    assert discovered[0]._smbus.name == 'i2c-42'


def test_ignores_non_bus_sysfs_entries(emulated_smbus, tmpdir):
    i2c_root = tmpdir.mkdir('sys').mkdir('bus').mkdir('i2c')
    devices = i2c_root.mkdir('devices')
    devices.mkdir('i2c-0')
    devices.mkdir('0-0050')  # SPD info chip on i2c-0
    devices.mkdir('i2c-DELL0829:00')  # i2c HID chip from Dell laptop

    virtual_bus = LinuxI2c(i2c_root=i2c_root)

    discovered = Canary.find_supported_devices(root_bus=virtual_bus)
    assert len(discovered) == 1
    assert discovered[0]._smbus.name == 'i2c-0'


def test_honors_a_bus_filter(emulated_smbus, tmpdir):
    i2c_root = tmpdir.mkdir('sys').mkdir('bus').mkdir('i2c')
    devices = i2c_root.mkdir('devices')
    devices.mkdir('i2c-0')
    devices.mkdir('i2c-1')

    virtual_bus = LinuxI2c(i2c_root=i2c_root)

    discovered = Canary.find_supported_devices(bus='i2c-1',
                                               root_bus=virtual_bus)
    assert len(discovered) == 1
    assert discovered[0]._smbus.name == 'i2c-1'


@pytest.fixture
def emulated_device(tmpdir, emulated_smbus):
    i2c_dev = Path(tmpdir.mkdir('i2c-0'))
    bus = LinuxI2cBus(i2c_dev=i2c_dev)
    dev = SmbusDriver(smbus=bus, description='Test', vendor_id=-1,
                      product_id=-1, address=-1)

    return (bus, dev)


def test_connect_is_unsafe(emulated_device):
    bus, dev = emulated_device

    def mock_open():
        nonlocal opened
        opened = True

    bus.open = mock_open
    opened = False

    dev.connect()
    assert not opened


def test_connects(emulated_device):
    bus, dev = emulated_device

    def mock_open():
        nonlocal opened
        opened = True

    bus.open = mock_open
    opened = False

    with dev.connect(unsafe='smbus') as cm:
        assert cm == dev
        assert opened


def test_loading_unnavailable_eeprom_returns_none(emulated_device):
    bus, dev = emulated_device
    assert bus.load_eeprom(0x51) is None


def test_loads_eeprom(emulated_device):
    bus, dev = emulated_device

    spd = bus._i2c_dev.joinpath('0-0051')
    spd.mkdir()
    spd.joinpath('eeprom').write_bytes(b'012345')
    spd.joinpath('name').write_text('name\n')

    assert bus.load_eeprom(0x51) == ('name', b'012345')



"""Test backward compatibility with liquidctl 1.4.x."""

import pytest
from _testutils import MockHidapiDevice

RADICAL_RED = [0xff, 0x35, 0x5e]
MOUNTAIN_MEADOW = [0x1a, 0xb3, 0x85]


def test_find_from_driver_package_still_available():
    from liquidctl.driver import find_liquidctl_devices


def test_kraken2_backwards_modes_are_deprecated(caplog):
    modes = ['backwards-spectrum-wave', 'backwards-marquee-3',
             'backwards-marquee-4', 'backwards-marquee-5',
             'backwards-marquee-6', 'covering-backwards-marquee',
             'backwards-moving-alternating', 'backwards-super-wave']

    from liquidctl.driver.kraken2 import Kraken2

    for mode in modes:
        base_mode = mode.replace('backwards-', '')

        old = Kraken2(MockHidapiDevice(), 'Mock X62',
                      device_type=Kraken2.DEVICE_KRAKENX)
        new = Kraken2(MockHidapiDevice(), 'Mock X62',
                      device_type=Kraken2.DEVICE_KRAKENX)

        colors = [RADICAL_RED, MOUNTAIN_MEADOW]

        old.set_color('ring', mode, colors)
        new.set_color('ring', base_mode, colors, direction='backward')

        assert old.device.sent == new.device.sent, \
               f'{mode} != {base_mode} + direction=backward'

        assert 'deprecated mode' in caplog.text


def test_kraken3_backwards_modes_are_deprecated(caplog):
    modes = ['backwards-spectrum-wave', 'backwards-marquee-3',
             'backwards-marquee-4', 'backwards-marquee-5',
             'backwards-marquee-6', 'backwards-moving-alternating-3',
             'covering-backwards-marquee', 'backwards-moving-alternating-4',
             'backwards-moving-alternating-5', 'backwards-moving-alternating-6',
             'backwards-rainbow-flow', 'backwards-super-rainbow',
             'backwards-rainbow-pulse']

    from liquidctl.driver.kraken3 import KrakenX3
    from liquidctl.driver.kraken3 import _COLOR_CHANNELS_KRAKENX
    from liquidctl.driver.kraken3 import _SPEED_CHANNELS_KRAKENX
    from liquidctl.driver.kraken3 import _HWMON_CTRL_MAPPING_KRAKENX

    for mode in modes:
        base_mode = mode.replace('backwards-', '')

        old = KrakenX3(MockHidapiDevice(), 'Mock X63',
                       speed_channels=_SPEED_CHANNELS_KRAKENX,
                       color_channels=_COLOR_CHANNELS_KRAKENX,
                       hwmon_ctrl_mapping=_HWMON_CTRL_MAPPING_KRAKENX)
        new = KrakenX3(MockHidapiDevice(), 'Mock X63',
                       speed_channels=_SPEED_CHANNELS_KRAKENX,
                       color_channels=_COLOR_CHANNELS_KRAKENX,
                       hwmon_ctrl_mapping=_HWMON_CTRL_MAPPING_KRAKENX)

        colors = [RADICAL_RED, MOUNTAIN_MEADOW]

        old.set_color('ring', mode, colors)
        new.set_color('ring', base_mode, colors, direction='backward')

        assert old.device.sent == new.device.sent, \
               f'{mode} != {base_mode} + direction=backward'

        assert 'deprecated mode' in caplog.text


def test_smart_device_v1_backwards_modes_are_deprecated(caplog):
    modes = ['backwards-spectrum-wave', 'backwards-marquee-3',
             'backwards-marquee-4', 'backwards-marquee-5',
             'backwards-marquee-6', 'covering-backwards-marquee',
             'backwards-moving-alternating', 'backwards-super-wave']

    from liquidctl.driver.smart_device import SmartDevice

    for mode in modes:
        base_mode = mode.replace('backwards-', '')

        old = SmartDevice(MockHidapiDevice(), 'Mock Smart Device',
                          speed_channel_count=3, color_channel_count=1)
        new = SmartDevice(MockHidapiDevice(), 'Mock Smart Device',
                          speed_channel_count=3, color_channel_count=1)

        colors = [RADICAL_RED, MOUNTAIN_MEADOW]

        old.set_color('led', mode, colors)
        new.set_color('led', base_mode, colors, direction='backward')

        assert old.device.sent == new.device.sent, \
               f'{mode} != {base_mode} + direction=backward'

        assert 'deprecated mode' in caplog.text


def test_hue2_backwards_modes_are_deprecated(caplog):
    modes = ['backwards-spectrum-wave', 'backwards-marquee-3',
             'backwards-marquee-4', 'backwards-marquee-5',
             'backwards-marquee-6', 'backwards-moving-alternating-3',
             'covering-backwards-marquee', 'backwards-moving-alternating-4',
             'backwards-moving-alternating-5', 'backwards-moving-alternating-6',
             'backwards-rainbow-flow', 'backwards-super-rainbow',
             'backwards-rainbow-pulse']

    from liquidctl.driver.smart_device import SmartDevice2

    for mode in modes:
        base_mode = mode.replace('backwards-', '')

        old = SmartDevice2(MockHidapiDevice(), 'Mock Smart Device V2',
                           speed_channel_count=3, color_channel_count=2)
        new = SmartDevice2(MockHidapiDevice(), 'Mock Smart Device V2',
                           speed_channel_count=3, color_channel_count=2)

        colors = [RADICAL_RED, MOUNTAIN_MEADOW]

        old.set_color('led1', mode, colors)
        new.set_color('led1', base_mode, colors, direction='backward')

        assert old.device.sent == new.device.sent, \
               f'{mode} != {base_mode} + direction=backward'

        assert 'deprecated mode' in caplog.text



"""Test backward compatibility with liquidctl 1.5.x."""

import pytest

from liquidctl.driver.hydro_platinum import HydroPlatinum
from liquidctl.driver.usb import hid


def test_matches_platinum_and_pro_xt_coolers_regardless_of_hydro(monkeypatch):
    mock_match = [
        {'vendor_id': 0x1b1c, 'product_id': 0x0c18},  # H100i Platinum
        {'vendor_id': 0x1b1c, 'product_id': 0x0c19},  # H100i Platinum SE
        {'vendor_id': 0x1b1c, 'product_id': 0x0c17},  # H115i Platinum
        {'vendor_id': 0x1b1c, 'product_id': 0x0c29},  # H60i Pro XT
        {'vendor_id': 0x1b1c, 'product_id': 0x0c20},  # H100i Pro XT
        {'vendor_id': 0x1b1c, 'product_id': 0x0c21},  # H115i Pro XT
        {'vendor_id': 0x1b1c, 'product_id': 0x0c22},  # H150i Pro XT
    ]

    mock_skip = [
        {'vendor_id': 0x1b1c, 'product_id': 0xffff},  # nothing
        {'vendor_id': 0xffff, 'product_id': 0x0c22},  # nothing
    ]

    mock_hids = mock_match + mock_skip

    for info in mock_hids:
        info.setdefault('path', b'')
        info.setdefault('path', b'')
        info.setdefault('serial_number', '')
        info.setdefault('release_number', 0)
        info.setdefault('manufacturer_string', '')
        info.setdefault('product_string', '')
        info.setdefault('usage_page', 0)
        info.setdefault('usage', 0)
        info.setdefault('interface_number', 0)

    def mock_enumerate(vid=0, pid=0):
        return mock_hids

    monkeypatch.setattr(hid, 'enumerate', mock_enumerate)

    def find(match):
        return HydroPlatinum.find_supported_devices(match=match)

    assert len(find('corsair hydro')) == 7
    assert len(find('hydro')) == 7
    assert len(find('corsair')) == 7

    assert len(find('corsair hydro h100i')) == 3
    assert len(find('hydro h100i')) == 3
    assert len(find('corsair h100i')) == 3



import pytest
from _testutils import VirtualSmbus

from liquidctl.driver.ddr4 import *
from liquidctl.error import *


# SPD samples


def patch_spd_dump(spd_dump, slice, new):
    spd_dump = bytearray(spd_dump)
    spd_dump[slice] = new
    return bytes(spd_dump)


_VENGEANCE_RGB_SAMPLE = bytes.fromhex(
    '23100c028521000800000003090300000000080cfc0300006c6c6c110874f00a'
    '2008000500a81e2b2b0000000000000000000000000000000000000016361636'
    '1636163600002b0c2b0c2b0c2b0c000000000000000000000000000000000000'
    '000000000000000000000000000000000000000000edb5ce0000000000c24da7'
    '1111010100000000000000000000000000000000000000000000000000000000'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '000000000000000000000000000000000000000000000000000000000000de27'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '029e00000000000000434d5233324758344d32433333333343313620200080ce'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '0c4a01200000000000a3000005fc3f04004d575710ac03f00a2008000500b022'
    '2c00000000000000009cceb5b5b5e7e700000000000000000000000000000000'
    '0000000000000000000000000000000000000000000000000000000000000000'
    '0000000000000000000000000000000000000000000000000000000000000000'
)

# clear the part number; the Vengeance RGB sample already didn't set the TS bit
_NON_TS_SPD = patch_spd_dump(_VENGEANCE_RGB_SAMPLE, slice(0x149, 0x15d), b' ' * 20)

# set the TS bit
_TS_SPD = patch_spd_dump(_NON_TS_SPD, 0x0e, 0x80)


# DDR4 SPD decoding


@pytest.fixture
def cmr_spd():
    return Ddr4Spd(_VENGEANCE_RGB_SAMPLE)


def test_spd_bytes_used(cmr_spd):
    assert cmr_spd.spd_bytes_used == 384


def test_spd_bytes_total(cmr_spd):
    assert cmr_spd.spd_bytes_total == 512


def test_spd_revision(cmr_spd):
    assert cmr_spd.spd_revision == (1, 0)


def test_dram_device_type(cmr_spd):
    assert cmr_spd.dram_device_type == Ddr4Spd.DramDeviceType.DDR4_SDRAM


def test_module_type(cmr_spd):
    assert cmr_spd.module_type == (Ddr4Spd.BaseModuleType.UDIMM, None)


def test_module_thermal_sensor(cmr_spd):
    assert not cmr_spd.module_thermal_sensor


def test_module_manufacturer(cmr_spd):
    assert cmr_spd.module_manufacturer == 'Corsair'


def test_module_part_number(cmr_spd):
    assert cmr_spd.module_part_number == 'CMR32GX4M2C3333C16'


def test_dram_manufacturer(cmr_spd):
    assert cmr_spd.dram_manufacturer == 'Samsung'


@pytest.fixture
def smbus():
    smbus = VirtualSmbus(parent_driver='i801_smbus')

    # hack: clear all spd addresses
    for address in range(0x50, 0x58):
        smbus._data[address] = None

    return smbus


# DDR4 modules using a TSE2004-compatible SPD EEPROM with temperature sensor


def test_tse2004_ignores_not_allowed_buses(smbus, monkeypatch):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _TS_SPD)

    checks = [
        ('parent_driver', 'other'),
    ]

    for attr, val in checks:
        with monkeypatch.context() as m:
            m.setattr(smbus, attr, val)
            assert list(Ddr4Temperature.probe(smbus)) == [], \
                f'changing {attr} did not cause a mismatch'


def test_tse2004_does_not_match_non_ee1004_device(smbus):
    smbus.emulate_eeprom_at(0x51, 'other', _TS_SPD)
    assert list(map(type, Ddr4Temperature.probe(smbus))) == []


def test_tse2004_does_not_match_non_ts_devices(smbus):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _NON_TS_SPD)
    assert list(map(type, Ddr4Temperature.probe(smbus))) == []


def test_tse2004_finds_ts_devices(smbus):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _TS_SPD)
    smbus.emulate_eeprom_at(0x53, 'ee1004', _TS_SPD)
    smbus.emulate_eeprom_at(0x55, 'ee1004', _TS_SPD)
    smbus.emulate_eeprom_at(0x57, 'ee1004', _TS_SPD)

    devs = list(Ddr4Temperature.probe(smbus))

    assert list(map(type, devs)) == [Ddr4Temperature] * 4
    assert devs[1].description.startswith('Corsair DIMM4')


def test_tse2004_get_status_is_unsafe(smbus):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _TS_SPD)
    dimm = next(Ddr4Temperature.probe(smbus))
    assert dimm.get_status() == []


def test_tse2004_get_status_reads_temperature(smbus):
    enable = ['smbus', 'ddr4_temperature']
    smbus.emulate_eeprom_at(0x51, 'ee1004', _TS_SPD)
    dimm = next(Ddr4Temperature.probe(smbus))

    with dimm.connect(unsafe=enable):
        smbus.write_word_data(0x19, 0x05, 0x9ce1)

        status = dimm.get_status(unsafe=enable)
        expected = [
            ('Temperature', 25.75, '°C'),
        ]

        assert status == expected


def test_tse2004_get_status_reads_negative_temperature(smbus):
    enable = ['smbus', 'ddr4_temperature']
    smbus.emulate_eeprom_at(0x51, 'ee1004', _TS_SPD)
    dimm = next(Ddr4Temperature.probe(smbus))

    with dimm.connect(unsafe=enable):
        smbus.write_word_data(0x19, 0x05, 0x741e)

        status = dimm.get_status(unsafe=enable)
        expected = [
            ('Temperature', -24.75, '°C'),
        ]

        assert status == expected


# Corsair Vengeance RGB


@pytest.fixture
def vengeance_rgb(smbus):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _VENGEANCE_RGB_SAMPLE)
    dimm = next(VengeanceRgb.probe(smbus))

    smbus.open()
    for reg in range(256):
        smbus.write_byte_data(0x59, reg, 0xba)
    smbus.close()

    return (smbus, dimm)


def test_vengeance_rgb_finds_devices(smbus):
    smbus.emulate_eeprom_at(0x51, 'ee1004', _VENGEANCE_RGB_SAMPLE)
    smbus.emulate_eeprom_at(0x53, 'ee1004', _VENGEANCE_RGB_SAMPLE)
    smbus.emulate_eeprom_at(0x55, 'ee1004', _VENGEANCE_RGB_SAMPLE)
    smbus.emulate_eeprom_at(0x57, 'ee1004', _VENGEANCE_RGB_SAMPLE)

    devs = list(VengeanceRgb.probe(smbus))

    assert list(map(type, devs)) == [VengeanceRgb] * 4
    assert devs[1].description.startswith('Corsair Vengeance RGB DIMM4')


def test_vengeance_get_status_reads_temperature(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    def forbid(*args, **kwargs):
        assert False, 'should not reach here'

    smbus.read_block_data = forbid

    with dimm.connect(unsafe=enable):
        smbus.write_word_data(0x19, 0x05, 0x9ce1)

        status = dimm.get_status(unsafe=enable)
        expected = [
            ('Temperature', 25.75, '°C'),
        ]

        assert status == expected


def test_vengeance_rgb_set_color_is_unsafe(vengeance_rgb):
    _, dimm = vengeance_rgb

    with pytest.raises(UnsafeFeaturesNotEnabled):
        assert dimm.set_color('led', 'off', [])

    with pytest.raises(UnsafeFeaturesNotEnabled):
        assert dimm.set_color('led', 'off', [], unsafe='vengeance_rgb')

    with pytest.raises(UnsafeFeaturesNotEnabled):
        assert dimm.set_color('led', 'off', [], unsafe='smbus')


def test_vengeance_rgb_asserts_rgb_address_validity(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        smbus.write_byte_data(0x59, 0xa6, 0x00)

        with pytest.raises(ExpectationNotMet):
            dimm.set_color('led', 'off', [], unsafe=enable)


def test_vengeance_rgb_sets_color_to_off(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        dimm.set_color('led', 'off', [], unsafe=enable)

        assert smbus.read_byte_data(0x59, 0xa4) == 0x00
        assert smbus.read_byte_data(0x59, 0xa6) == 0x00
        assert smbus.read_byte_data(0x59, 0xa7) == 0x01

        for color_component in range(0xb0, 0xb3):
            assert smbus.read_byte_data(0x59, color_component) == 0x00


def test_vengeance_rgb_sets_color_to_fixed(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        dimm.set_color('led', 'fixed', [radical_red], unsafe=enable)

        assert smbus.read_byte_data(0x59, 0xa4) == 0x00
        assert smbus.read_byte_data(0x59, 0xa6) == 0x00
        assert smbus.read_byte_data(0x59, 0xa7) == 0x01

        assert smbus.read_byte_data(0x59, 0xb0) == 0xff
        assert smbus.read_byte_data(0x59, 0xb1) == 0x35
        assert smbus.read_byte_data(0x59, 0xb2) == 0x5e


def test_vengeance_rgb_sets_color_to_breathing(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        mountain_meadow = [0x1a, 0xb3, 0x85]
        dimm.set_color('led', 'breathing', [radical_red, mountain_meadow],
                       unsafe=enable)

        assert smbus.read_byte_data(0x59, 0xa4) == 0x20
        assert smbus.read_byte_data(0x59, 0xa5) == 0x20
        assert smbus.read_byte_data(0x59, 0xa6) == 0x02
        assert smbus.read_byte_data(0x59, 0xa7) == 0x02

        assert smbus.read_byte_data(0x59, 0xb0) == 0xff
        assert smbus.read_byte_data(0x59, 0xb1) == 0x35
        assert smbus.read_byte_data(0x59, 0xb2) == 0x5e

        assert smbus.read_byte_data(0x59, 0xb3) == 0x1a
        assert smbus.read_byte_data(0x59, 0xb4) == 0xb3
        assert smbus.read_byte_data(0x59, 0xb5) == 0x85


def test_vengeance_rgb_sets_single_color_to_breathing(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        dimm.set_color('led', 'breathing', [radical_red],
                       unsafe=enable)

        assert smbus.read_byte_data(0x59, 0xa4) == 0x20
        assert smbus.read_byte_data(0x59, 0xa5) == 0x20
        assert smbus.read_byte_data(0x59, 0xa6) == 0x00  # special case
        assert smbus.read_byte_data(0x59, 0xa7) == 0x01

        assert smbus.read_byte_data(0x59, 0xb0) == 0xff
        assert smbus.read_byte_data(0x59, 0xb1) == 0x35
        assert smbus.read_byte_data(0x59, 0xb2) == 0x5e


def test_vengeance_rgb_sets_color_to_fading(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        mountain_meadow = [0x1a, 0xb3, 0x85]
        dimm.set_color('led', 'fading', [radical_red, mountain_meadow],
                       unsafe=enable)

        assert smbus.read_byte_data(0x59, 0xa4) == 0x20
        assert smbus.read_byte_data(0x59, 0xa5) == 0x20
        assert smbus.read_byte_data(0x59, 0xa6) == 0x01
        assert smbus.read_byte_data(0x59, 0xa7) == 0x02

        assert smbus.read_byte_data(0x59, 0xb0) == 0xff
        assert smbus.read_byte_data(0x59, 0xb1) == 0x35
        assert smbus.read_byte_data(0x59, 0xb2) == 0x5e

        assert smbus.read_byte_data(0x59, 0xb3) == 0x1a
        assert smbus.read_byte_data(0x59, 0xb4) == 0xb3
        assert smbus.read_byte_data(0x59, 0xb5) == 0x85


def test_vengeance_rgb_animation_speed_presets_set_correct_timings(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        mountain_meadow = [0x1a, 0xb3, 0x85]

        presets = ['slowest', 'slower', 'normal', 'faster', 'fastest']
        timings = [0x3f, 0x30, 0x20, 0x10, 0x01]
        for preset, timing in zip(presets, timings):
            dimm.set_color('led', 'fading', [radical_red, mountain_meadow],
                           speed=preset, unsafe=enable)

            assert smbus.read_byte_data(0x59, 0xa4) == timing, f'wrong tp1 for {preset!r}'
            assert smbus.read_byte_data(0x59, 0xa5) == timing, f'wrong tp2 for {preset!r}'


def test_vengeance_rgb_animation_transition_ticks_overrides_tp1(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        mountain_meadow = [0x1a, 0xb3, 0x85]
        dimm.set_color('led', 'fading', [radical_red, mountain_meadow],
                       transition_ticks=0x01, unsafe=enable)
        assert smbus.read_byte_data(0x59, 0xa4) == 0x01
        assert smbus.read_byte_data(0x59, 0xa5) == 0x20


def test_vengeance_rgb_animation_transition_ticks_overrides_tp2(vengeance_rgb):
    enable = ['smbus', 'vengeance_rgb']
    smbus, dimm = vengeance_rgb

    with dimm.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        mountain_meadow = [0x1a, 0xb3, 0x85]
        dimm.set_color('led', 'fading', [radical_red, mountain_meadow],
                       stable_ticks=0x01, unsafe=enable)
        assert smbus.read_byte_data(0x59, 0xa4) == 0x20
        assert smbus.read_byte_data(0x59, 0xa5) == 0x01



"""liquidctl drivers for fourth-generation NZXT Kraken X liquid coolers.

Supported devices:

- NZXT Kraken X (X53, X63 and X73)
- NZXT Kraken Z (Z53, Z63 and Z73)

Copyright Tom Frey, Jonas Malaco, Shady Nawara and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""
# uses the psf/black style

import itertools
import io
import math
import logging
import sys
import time

from PIL import Image, ImageSequence

if sys.platform == "win32":
    from winusbcdc import WinUsbPy

from liquidctl.driver.usb import PyUsbDevice, UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import (
    LazyHexRepr,
    normalize_profile,
    interpolate_profile,
    clamp,
    Hue2Accessory,
    HUE2_MAX_ACCESSORIES_IN_CHANNEL,
    map_direction,
)

_LOGGER = logging.getLogger(__name__)

_READ_LENGTH = 64
_WRITE_LENGTH = 64
_MAX_READ_ATTEMPTS = 12
_BULK_WRITE_LENGTH = 512

_LCD_TOTAL_MEMORY = 24320

_STATUS_TEMPERATURE = "Liquid temperature"
_STATUS_PUMP_SPEED = "Pump speed"
_STATUS_PUMP_DUTY = "Pump duty"
_STATUS_FAN_SPEED = "Fan speed"
_STATUS_FAN_DUTY = "Fan duty"

# Available speed channels for model X coolers
# name -> (channel_id, min_duty, max_duty)
# TODO adjust min duty value to what the firmware enforces
_SPEED_CHANNELS_KRAKENX = {
    "pump": (0x1, 20, 100),
}

# Available speed channels for model Z coolers
# name -> (channel_id, min_duty, max_duty)
# TODO adjust min duty values to what the firmware enforces
_SPEED_CHANNELS_KRAKENZ = {
    "pump": (0x1, 20, 100),
    "fan": (0x2, 0, 100),
}

_CRITICAL_TEMPERATURE = 59

# Available color channels and IDs for model X coolers
_COLOR_CHANNELS_KRAKENX = {"external": 0b001, "ring": 0b010, "logo": 0b100, "sync": 0b111}

# Available color channels and IDs for model Z coolers
_COLOR_CHANNELS_KRAKENZ = {
    "external": 0b001,
}

_HWMON_CTRL_MAPPING_KRAKENX = {"pump": 1}

_HWMON_CTRL_MAPPING_KRAKENZ = {"pump": 1, "fan": 2}

# Available LED channel modes/animations
# name -> (mode, size/variant, speed scale, min colors, max colors)
# FIXME any point in a one-color *alternating* or tai-chi animations?
# FIXME are all modes really supported by all channels? (this is better because
#       of synchronization, but it's not how the previous generation worked, so
#       I would like to double check)
_COLOR_MODES = {
    "off": (0x00, 0x00, 0, 0, 0),
    "fixed": (0x00, 0x00, 0, 1, 1),
    "fading": (0x01, 0x00, 1, 1, 8),
    "super-fixed": (0x01, 0x01, 9, 1, 40),
    "spectrum-wave": (0x02, 0x00, 2, 0, 0),
    "marquee-3": (0x03, 0x03, 2, 1, 1),
    "marquee-4": (0x03, 0x04, 2, 1, 1),
    "marquee-5": (0x03, 0x05, 2, 1, 1),
    "marquee-6": (0x03, 0x06, 2, 1, 1),
    "covering-marquee": (0x04, 0x00, 2, 1, 8),
    "alternating-3": (0x05, 0x03, 3, 1, 2),
    "alternating-4": (0x05, 0x04, 3, 1, 2),
    "alternating-5": (0x05, 0x05, 3, 1, 2),
    "alternating-6": (0x05, 0x06, 3, 1, 2),
    "moving-alternating-3": (0x05, 0x03, 4, 1, 2),
    "moving-alternating-4": (0x05, 0x04, 4, 1, 2),
    "moving-alternating-5": (0x05, 0x05, 4, 1, 2),
    "moving-alternating-6": (0x05, 0x06, 4, 1, 2),
    "pulse": (0x06, 0x00, 5, 1, 8),
    "breathing": (0x07, 0x00, 6, 1, 8),
    "super-breathing": (0x03, 0x00, 10, 1, 40),
    "candle": (0x08, 0x00, 0, 1, 1),
    "starry-night": (0x09, 0x00, 5, 1, 1),
    "rainbow-flow": (0x0B, 0x00, 2, 0, 0),
    "super-rainbow": (0x0C, 0x00, 2, 0, 0),
    "rainbow-pulse": (0x0D, 0x00, 2, 0, 0),
    "loading": (0x10, 0x00, 8, 1, 1),
    "tai-chi": (0x0E, 0x00, 7, 1, 2),
    "water-cooler": (0x0F, 0x00, 6, 2, 2),
    "wings": (None, 0x00, 11, 1, 1),
    # deprecated in favor of direction=backward
    "backwards-spectrum-wave": (0x02, 0x00, 2, 0, 0),
    "backwards-marquee-3": (0x03, 0x03, 2, 1, 1),
    "backwards-marquee-4": (0x03, 0x04, 2, 1, 1),
    "backwards-marquee-5": (0x03, 0x05, 2, 1, 1),
    "backwards-marquee-6": (0x03, 0x06, 2, 1, 1),
    "covering-backwards-marquee": (0x04, 0x00, 2, 1, 8),
    "backwards-moving-alternating-3": (0x05, 0x03, 4, 1, 2),
    "backwards-moving-alternating-4": (0x05, 0x04, 4, 1, 2),
    "backwards-moving-alternating-5": (0x05, 0x05, 4, 1, 2),
    "backwards-moving-alternating-6": (0x05, 0x06, 4, 1, 2),
    "backwards-rainbow-flow": (0x0B, 0x00, 2, 0, 0),
    "backwards-super-rainbow": (0x0C, 0x00, 2, 0, 0),
    "backwards-rainbow-pulse": (0x0D, 0x00, 2, 0, 0),
}

# A static value per channel that is somehow related to animation time and
# synchronization, although the specific mechanism is not yet understood.
# Could require information from `initialize`, but more testing is required.
_STATIC_VALUE = {
    0b001: 40,  # may result in long all-off intervals (FIXME?)
    0b010: 8,
    0b100: 1,
    0b111: 40,  # may result in long all-off intervals (FIXME?)
}

# Speed scale/timing bytes
# scale -> (slowest, slower, normal, faster, fastest)
_SPEED_VALUE = {
    0: ([0x32, 0x00], [0x32, 0x00], [0x32, 0x00], [0x32, 0x00], [0x32, 0x00]),
    1: ([0x50, 0x00], [0x3C, 0x00], [0x28, 0x00], [0x14, 0x00], [0x0A, 0x00]),
    2: ([0x5E, 0x01], [0x2C, 0x01], [0xFA, 0x00], [0x96, 0x00], [0x50, 0x00]),
    3: ([0x40, 0x06], [0x14, 0x05], [0xE8, 0x03], [0x20, 0x03], [0x58, 0x02]),
    4: ([0x20, 0x03], [0xBC, 0x02], [0xF4, 0x01], [0x90, 0x01], [0x2C, 0x01]),
    5: ([0x19, 0x00], [0x14, 0x00], [0x0F, 0x00], [0x07, 0x00], [0x04, 0x00]),
    6: ([0x28, 0x00], [0x1E, 0x00], [0x14, 0x00], [0x0A, 0x00], [0x04, 0x00]),
    7: ([0x32, 0x00], [0x28, 0x00], [0x1E, 0x00], [0x14, 0x00], [0x0A, 0x00]),
    8: ([0x14, 0x00], [0x14, 0x00], [0x14, 0x00], [0x14, 0x00], [0x14, 0x00]),
    9: ([0x00, 0x00], [0x00, 0x00], [0x00, 0x00], [0x00, 0x00], [0x00, 0x00]),
    10: ([0x37, 0x00], [0x28, 0x00], [0x19, 0x00], [0x0A, 0x00], [0x00, 0x00]),
    11: ([0x6E, 0x00], [0x53, 0x00], [0x39, 0x00], [0x2E, 0x00], [0x20, 0x00]),
}

_ANIMATION_SPEEDS = {
    "slowest": 0x0,
    "slower": 0x1,
    "normal": 0x2,
    "faster": 0x3,
    "fastest": 0x4,
}


class KrakenX3(UsbHidDriver):
    """Fourth-generation Kraken X liquid cooler."""

    # support for hwmon: nzxt-kraken3, liquidtux
    # https://github.com/liquidctl/liquidtux/blob/3b80dafead6f/nzxt-kraken3.c

    _MATCHES = [
        (
            0x1E71,
            0x2007,
            "NZXT Kraken X (X53, X63 or X73)",
            {
                "speed_channels": _SPEED_CHANNELS_KRAKENX,
                "color_channels": _COLOR_CHANNELS_KRAKENX,
                "hwmon_ctrl_mapping": _HWMON_CTRL_MAPPING_KRAKENX,
            },
        ),
        (
            0x1E71,
            0x2014,
            "NZXT Kraken X (X53, X63 or X73)",
            {
                "speed_channels": _SPEED_CHANNELS_KRAKENX,
                "color_channels": _COLOR_CHANNELS_KRAKENX,
                "hwmon_ctrl_mapping": _HWMON_CTRL_MAPPING_KRAKENX,
            },
        ),
    ]

    def __init__(
        self, device, description, speed_channels, color_channels, hwmon_ctrl_mapping, **kwargs
    ):
        super().__init__(device, description)
        self._speed_channels = speed_channels
        self._color_channels = color_channels
        self._hwmon_ctrl_mapping = hwmon_ctrl_mapping

    def initialize(self, direct_access=False, **kwargs):
        """Initialize the device and the driver.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        self.device.clear_enqueued_reports()
        # request static infos
        self._write([0x10, 0x01])  # firmware info
        self._write([0x20, 0x03])  # lighting info

        # initialize
        if self._hwmon and not direct_access:
            _LOGGER.info(
                "bound to %s kernel driver, assuming it is already initialized", self._hwmon.driver
            )
        else:
            if self._hwmon:
                _LOGGER.warning(
                    "forcing re-initialization despite %s kernel driver", self._hwmon.driver
                )
            update_interval = (lambda secs: 1 + round((secs - 0.5) / 0.25))(0.5)  # see issue #128
            self._write([0x70, 0x02, 0x01, 0xB8, update_interval])
            self._write([0x70, 0x01])

        self._status = []

        self._read_until({b"\x11\x01": self.parse_firm_info, b"\x21\x03": self.parse_led_info})
        return sorted(self._status)

    def parse_firm_info(self, msg):
        fw = f"{msg[0x11]}.{msg[0x12]}.{msg[0x13]}"
        self._status.append(("Firmware version", fw, ""))

    def parse_led_info(self, msg):
        channel_count = msg[14]
        assert channel_count == len(self._color_channels) - (
            "sync" in self._color_channels
        ), f"Unexpected number of color channels received: {channel_count}"

        def find(channel, accessory):
            offset = 15  # offset of first channel/first accessory
            acc_id = msg[offset + channel * HUE2_MAX_ACCESSORIES_IN_CHANNEL + accessory]
            return Hue2Accessory(acc_id) if acc_id else None

        for i in range(HUE2_MAX_ACCESSORIES_IN_CHANNEL):
            accessory = find(0, i)
            if not accessory:
                break
            self._status.append((f"LED accessory {i + 1}", accessory, ""))

        if len(self._color_channels) > 1:
            found_ring = find(1, 0) == Hue2Accessory.KRAKENX_GEN4_RING
            found_logo = find(2, 0) == Hue2Accessory.KRAKENX_GEN4_LOGO
            self._status.append(("Pump Ring LEDs", "detected" if found_ring else "missing", ""))
            self._status.append(("Pump Logo LEDs", "detected" if found_logo else "missing", ""))
            assert found_ring and found_logo, "Pump ring and/or logo were not detected"

    def _get_status_directly(self):
        self.device.clear_enqueued_reports()
        msg = self._read()
        if msg[15:17] == [0xFF, 0xFF]:
            _LOGGER.warning("unexpected temperature reading, possible firmware fault;")
            _LOGGER.warning("try resetting the device or updating the firmware")
            _LOGGER.warning("(see https://github.com/liquidctl/liquidctl/issues/172)")
        return [
            (_STATUS_TEMPERATURE, msg[15] + msg[16] / 10, "°C"),
            (_STATUS_PUMP_SPEED, msg[18] << 8 | msg[17], "rpm"),
            (_STATUS_PUMP_DUTY, msg[19], "%"),
        ]

    def _get_status_from_hwmon(self):
        status_readings = [
            (_STATUS_TEMPERATURE, self._hwmon.read_int("temp1_input") * 1e-3, "°C"),
            (_STATUS_PUMP_SPEED, self._hwmon.read_int("fan1_input"), "rpm"),
        ]

        if self._hwmon.has_attribute("pwm1"):
            status_readings.append(
                (_STATUS_PUMP_DUTY, self._hwmon.read_int("pwm1") * 100.0 / 255, "%")
            )
        else:
            # An older version of the kernel driver only exposed coolant temp and pump speed
            _LOGGER.warning("pump duty cannot be read from %s kernel driver", self._hwmon.driver)

        return status_readings

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        if self._hwmon and not direct_access:
            _LOGGER.info("bound to %s kernel driver, reading status from hwmon", self._hwmon.driver)
            return self._get_status_from_hwmon()

        if self._hwmon:
            _LOGGER.warning(
                "directly reading the status despite %s kernel driver", self._hwmon.driver
            )

        return self._get_status_directly()

    def set_color(self, channel, mode, colors, speed="normal", direction="forward", **kwargs):
        """Set the color mode for a specific channel."""

        if "backwards" in mode:
            _LOGGER.warning("deprecated mode, move to direction=backward option")
            mode = mode.replace("backwards-", "")
            direction = "backward"

        cid = self._color_channels[channel]
        _, _, _, mincolors, maxcolors = _COLOR_MODES[mode]
        colors = [[g, r, b] for [r, g, b] in colors]
        if len(colors) < mincolors:
            raise ValueError(f"not enough colors for mode={mode}, at least {mincolors} required")
        elif maxcolors == 0:
            if colors:
                _LOGGER.warning("too many colors for mode=%s, none needed", mode)
            colors = [[0, 0, 0]]  # discard the input but ensure at least one step
        elif len(colors) > maxcolors:
            _LOGGER.warning("too many colors for mode=%s, dropping to %d", mode, maxcolors)
            colors = colors[:maxcolors]

        sval = _ANIMATION_SPEEDS[speed]
        self._write_colors(cid, mode, colors, sval, direction)

    def _set_speed_profile_hwmon(self, channel, interp):
        hwmon_ctrl_channel = self._hwmon_ctrl_mapping[channel]

        # Write duty curve for channel
        for idx, duty in enumerate(interp):
            pwm_duty = duty * 255 // 100
            self._hwmon.write_int(f"temp{hwmon_ctrl_channel}_auto_point{idx + 1}_pwm", pwm_duty)

        # The device can get confused when hammered with HID reports, which can happen when
        # we set all curve points (done above) through the kernel driver, when the device
        # is in curve mode. In that case, the driver sends a report for each point value change
        # to update it. We send the whole curve to the device again by setting pwmX_enable to 2,
        # regardless of what it was, to ensure that the curve is properly applied. Wait just for
        # a bit to ensure that goes through
        time.sleep(0.2)

        # Set channel to curve mode
        self._hwmon.write_int(f"pwm{hwmon_ctrl_channel}_enable", 2)

    def set_speed_profile(self, channel, profile, direct_access=False, **kwargs):
        """Set channel to use a speed duty profile."""

        cid, dmin, dmax = self._speed_channels[channel]
        header = [0x72, cid, 0x00, 0x00]
        norm = normalize_profile(profile, _CRITICAL_TEMPERATURE)
        stdtemps = list(range(20, _CRITICAL_TEMPERATURE + 1))
        interp = [clamp(interpolate_profile(norm, t), dmin, dmax) for t in stdtemps]
        for temp, duty in zip(stdtemps, interp):
            _LOGGER.info(
                "setting %s PWM duty to %d%% for liquid temperature >= %d°C", channel, duty, temp
            )

        if self._hwmon:
            hwmon_pwm_enable_name = f"pwm{self._hwmon_ctrl_mapping[channel]}_enable"

            # Check if the required attribute is present
            if self._hwmon.has_attribute(hwmon_pwm_enable_name):
                # It is, and if we have to use direct access, warn that we are sidestepping the kernel driver
                if direct_access:
                    _LOGGER.warning(
                        "directly writing duty curve despite %s kernel driver having support",
                        self._hwmon.driver,
                    )
                    return self._write(header + interp)

                _LOGGER.info(
                    "bound to %s kernel driver, writing duty curve to hwmon", self._hwmon.driver
                )
                return self._set_speed_profile_hwmon(channel, interp)
            elif not direct_access:
                _LOGGER.warning(
                    "required duty curve functionality is not available in %s kernel driver, falling back to direct access",
                    self._hwmon.driver,
                )

        return self._write(header + interp)

    def _set_fixed_speed_directly(self, channel, duty):
        self.set_speed_profile(channel, [(0, duty), (_CRITICAL_TEMPERATURE - 1, duty)], True)

    def _set_fixed_speed_hwmon(self, channel, duty):
        hwmon_pwm_name = f"pwm{self._hwmon_ctrl_mapping[channel]}"
        hwmon_pwm_enable_name = f"{hwmon_pwm_name}_enable"

        # Convert duty from percent to PWM range (0-255)
        pwm_duty = duty * 255 // 100

        # Write duty to hwmon
        self._hwmon.write_int(hwmon_pwm_name, pwm_duty)

        # Set channel to direct percent mode
        self._hwmon.write_int(hwmon_pwm_enable_name, 1)

    def set_fixed_speed(self, channel, duty, direct_access=False, **kwargs):
        """Set channel to a fixed speed duty."""

        if self._hwmon:
            _, dmin, dmax = self._speed_channels[channel]
            duty = clamp(duty, dmin, dmax)

            hwmon_pwm_name = f"pwm{self._hwmon_ctrl_mapping[channel]}"

            # Check if the required attribute is present
            if self._hwmon.has_attribute(hwmon_pwm_name):
                # It is, and if we have to use direct access, warn that we are sidestepping the kernel driver
                if direct_access:
                    _LOGGER.warning(
                        "directly writing fixed speed despite %s kernel driver having support",
                        self._hwmon.driver,
                    )
                    return self._set_fixed_speed_directly(channel, duty)

                _LOGGER.info(
                    "bound to %s kernel driver, writing fixed speed to hwmon", self._hwmon.driver
                )
                return self._set_fixed_speed_hwmon(channel, duty)
            elif not direct_access:
                _LOGGER.warning(
                    "required PWM functionality is not available in %s kernel driver, falling back to direct access",
                    self._hwmon.driver,
                )

        return self._set_fixed_speed_directly(channel, duty)

    def _read(self):
        data = self.device.read(_READ_LENGTH)
        return data

    def _read_until(self, parsers):
        for _ in range(_MAX_READ_ATTEMPTS):
            msg = self._read()
            prefix = bytes(msg[0:2])
            func = parsers.pop(prefix, None)
            if func:
                func(msg)
            if not parsers:
                return
        assert False, f"missing messages (attempts={_MAX_READ_ATTEMPTS}, missing={len(parsers)})"

    def _write(self, data):
        padding = [0x0] * (_WRITE_LENGTH - len(data))
        self.device.write(data + padding)

    def _write_colors(self, cid, mode, colors, sval, direction):
        mval, size_variant, speed_scale, mincolors, maxcolors = _COLOR_MODES[mode]
        color_count = len(colors)

        if "super-fixed" == mode or "super-breathing" == mode:
            color = list(itertools.chain(*colors)) + [0x00, 0x00, 0x00] * (maxcolors - color_count)
            speed_value = _SPEED_VALUE[speed_scale][sval]
            self._write([0x22, 0x10, cid, 0x00] + color)
            self._write([0x22, 0x11, cid, 0x00])
            self._write(
                [0x22, 0xA0, cid, 0x00, mval]
                + speed_value
                + [0x08, 0x00, 0x00, 0x80, 0x00, 0x32, 0x00, 0x00, 0x01]
            )

        elif mode == "wings":  # wings requires special handling
            self._write([0x22, 0x10, cid])  # clear out all independent LEDs
            self._write([0x22, 0x11, cid])  # clear out all independent LEDs
            color_lists = {}
            color_lists[0] = colors[0] * 2
            color_lists[1] = [int(x // 2.5) for x in color_lists[0]]
            color_lists[2] = [int(x // 4) for x in color_lists[1]]
            color_lists[3] = [0x00] * 8
            speed_value = _SPEED_VALUE[speed_scale][sval]
            for i in range(8):  # send color scheme first, before enabling wings mode
                mod = 0x05 if i in [3, 7] else 0x01
                alt = [0x04, 0x84] if i // 4 == 0 else [0x84, 0x04]
                msg = (
                    [0x22, 0x20, cid, i, 0x04]
                    + speed_value
                    + [mod]
                    + [0x00] * 7
                    + [0x02]
                    + alt
                    + [0x00] * 10
                )
                self._write(msg + color_lists[i % 4])
            self._write([0x22, 0x03, cid, 0x08])  # this actually enables wings mode

        else:
            opcode = [0x2A, 0x04]
            address = [cid, cid]
            speed_value = _SPEED_VALUE[speed_scale][sval]
            header = opcode + address + [mval] + speed_value
            color = list(itertools.chain(*colors)) + [0, 0, 0] * (16 - color_count)

            if "marquee" in mode:
                backward_byte = 0x04
            elif mode == "starry-night" or "moving-alternating" in mode:
                backward_byte = 0x01
            else:
                backward_byte = 0x00

            backward_byte += map_direction(direction, 0, 0x02)

            if mode == "fading" or mode == "pulse" or mode == "breathing":
                mode_related = 0x08
            elif mode == "tai-chi":
                mode_related = 0x05
            elif mode == "water-cooler":
                mode_related = 0x05
                color_count = 0x01
            elif mode == "loading":
                mode_related = 0x04
            else:
                mode_related = 0x00

            static_byte = _STATIC_VALUE[cid]
            led_size = size_variant if mval == 0x03 or mval == 0x05 else 0x03
            footer = [backward_byte, color_count, mode_related, static_byte, led_size]
            self._write(header + color + footer)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


class KrakenZ3(KrakenX3):
    """Fourth-generation Kraken Z liquid cooler."""

    _MATCHES = [
        (
            0x1E71,
            0x3008,
            "NZXT Kraken Z (Z53, Z63 or Z73) (experimental)",
            {
                "speed_channels": _SPEED_CHANNELS_KRAKENZ,
                "color_channels": _COLOR_CHANNELS_KRAKENZ,
                "hwmon_ctrl_mapping": _HWMON_CTRL_MAPPING_KRAKENZ,
            },
        )
    ]

    def __init__(self, device, description, speed_channels, color_channels, **kwargs):
        super().__init__(device, description, speed_channels, color_channels, **kwargs)

        if sys.platform == "win32":
            self.bulk_device = WinUsbPy()
            found_device = self._find_winusb_device(
                "vid_1e71", "pid_3008", self.device.serial_number
            )
            if not found_device:
                self.bulk_device = None
        else:
            self.bulk_device = next(
                (
                    handle
                    for handle in PyUsbDevice.enumerate(self.vendor_id, self.product_id)
                    if handle.serial_number == self.device.serial_number
                ),
                None,
            )
            if self.bulk_device:
                self.bulk_device.open()

        self.orientation = 0  # 0 = Normal, 1 = +90 degrees, 2 = 180 degrees, 3 = -90(270) degrees
        self.brightness = 50  # default 50%

    def _find_winusb_device(self, vid, pid, serial):
        winusb_devices = self.bulk_device.list_usb_devices(
            deviceinterface=True, present=True, findparent=True
        )
        for device in winusb_devices:
            if (
                device.path.find(vid + "&" + pid) != -1
                and device.parent
                and device.parent.find(serial) != -1
            ):
                self.bulk_device.init_winusb_device_with_path(device.path)
                return True
        return False

    def initialize(self, direct_access=False, **kwargs):
        """Initialize the device and the driver.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        self.device.clear_enqueued_reports()
        # request static infos
        self._write([0x10, 0x01])  # firmware info
        self._write([0x20, 0x03])  # lighting info
        self._write([0x30, 0x01])  # lcd info

        # initialize
        if self._hwmon and not direct_access:
            _LOGGER.info(
                "bound to %s kernel driver, assuming it is already initialized", self._hwmon.driver
            )
        else:
            if self._hwmon:
                _LOGGER.warning(
                    "forcing re-initialization despite %s kernel driver", self._hwmon.driver
                )
        update_interval = (lambda secs: 1 + round((secs - 0.5) / 0.25))(0.5)  # see issue #128
        self._write([0x70, 0x02, 0x01, 0xB8, update_interval])
        self._write([0x70, 0x01])

        self._status = []

        self._read_until(
            {
                b"\x11\x01": self.parse_firm_info,
                b"\x21\x03": self.parse_led_info,
                b"\x31\x01": self.parse_lcd_info,
            }
        )

        return sorted(self._status)

    def parse_lcd_info(self, msg):
        self.brightness = msg[0x18]
        self.orientation = msg[0x1A]
        self._status.append(("LCD Brightness", self.brightness, "%"))
        self._status.append(("LCD Orientation", self.orientation * 90, "°"))

    def _get_status_directly(self):
        self.device.clear_enqueued_reports()
        self._write([0x74, 0x01])
        msg = self._read()
        if msg[15:17] == [0xFF, 0xFF]:
            _LOGGER.warning("unexpected temperature reading, possible firmware fault;")
            _LOGGER.warning("try resetting the device or updating the firmware")
        return [
            (_STATUS_TEMPERATURE, msg[15] + msg[16] / 10, "°C"),
            (_STATUS_PUMP_SPEED, msg[18] << 8 | msg[17], "rpm"),
            (_STATUS_PUMP_DUTY, msg[19], "%"),
            (_STATUS_FAN_SPEED, msg[24] << 8 | msg[23], "rpm"),
            (_STATUS_FAN_DUTY, msg[25], "%"),
        ]

    def _get_status_from_hwmon(self):
        return [
            (_STATUS_TEMPERATURE, self._hwmon.read_int("temp1_input") * 1e-3, "°C"),
            (_STATUS_PUMP_SPEED, self._hwmon.read_int("fan1_input"), "rpm"),
            (_STATUS_PUMP_DUTY, self._hwmon.read_int("pwm1") * 100.0 / 255, "%"),
            (_STATUS_FAN_SPEED, self._hwmon.read_int("fan2_input"), "rpm"),
            (_STATUS_FAN_DUTY, self._hwmon.read_int("pwm2") * 100.0 / 255, "%"),
        ]

    def _read_until_first_match(self, parsers):
        for _ in range(_MAX_READ_ATTEMPTS):
            msg = self._read()
            prefix = bytes(msg[0:2])
            func = parsers.pop(prefix, None)
            if func:
                return func(msg)
            if not parsers:
                return
        assert False, f"missing messages (attempts={_MAX_READ_ATTEMPTS}, missing={len(parsers)})"

    def _write_then_read(self, data):
        self._write(data)
        return self._read()

    def _bulk_write(self, data):
        padding = [0x0] * (_BULK_WRITE_LENGTH - len(data))
        out_data = data + padding
        if sys.platform == "win32":
            _LOGGER.debug("writing %d bytes: %r", len(out_data), LazyHexRepr(out_data))
            out_data = bytes(out_data)
        self.bulk_device.write(0x2, out_data)

    def set_screen(self, channel, mode, value, **kwargs):
        """Set the screen mode and content.

        Unstable.

        Supported channels, modes and values:

        | Channel | Mode | Value |
        | --- | --- | --- |
        | `lcd` | `liquid` | — |
        | `lcd` | `brightness` | int between `0` and `100` (%) |
        | `lcd` | `orientation` | `0`, `90`, `180` or `270` (°) |
        | `lcd` | `static` | path to image |
        | `lcd` | `gif` | path to animated GIF |
        """

        assert channel.lower() == "lcd", "Invalid Channel, valid: lcd, provided: " + channel
        assert mode != None, "No mode specified"

        if mode != "liquid":
            assert value != None, f"Mode: {mode} needs a value"

        # get orientation and brightness
        self._write([0x30, 0x01])

        def parse_lcd_info(msg):
            self.brightness = msg[0x18]
            self.orientation = msg[0x1A]

        self._read_until({b"\x31\x01": parse_lcd_info})

        if mode == "brightness":
            value_int = int(value)
            assert value_int >= 0 and value_int <= 100, "Invalid brightness value"
            self._write([0x30, 0x02, 0x01, value_int, 0x0, 0x0, 0x1, self.orientation])
            return
        elif mode == "orientation":
            value_int = int(value)
            assert (
                value_int == 0 or value_int == 90 or value_int == 180 or value_int == 270
            ), "Invalid orientation value"
            self._write([0x30, 0x02, 0x01, self.brightness, 0x0, 0x0, 0x1, int(value_int / 90)])
            return
        elif mode == "static":
            data = self._prepare_static_file(value, self.orientation)
            self._send_data(data, [0x02, 0x0, 0x0, 0x0, 0x0, 0x40, 0x06])
            return
        elif mode == "gif":
            data = self._prepare_gif_file(value, self.orientation)
            assert (
                len(data) / 1000 < _LCD_TOTAL_MEMORY
            ), f"Max file size after resize is 24MB, selected file is {len(data) / 1000000}MB"
            self._send_data(data, [0x01, 0x0, 0x0, 0x0] + list(len(data).to_bytes(3, "little")))
            return
        elif mode == "liquid":
            self._switch_bucket(0, 2)
            return

        if self.bulk_device and (mode == "static" or mode == "gif"):  # release device when finished
            if sys.platform == "win32":
                self.bulk_device.close_winusb_device()
            else:
                self.bulk_device.release()

        raise TypeError("Invalid mode")

    def _prepare_static_file(self, path, rotation):
        """
        path is the path to any image file
        Rotation is expected as 0 = no rotation, 1 = 90 degrees, 2 = 180 degrees, 3 = 270 degrees
        """
        img = Image.open(path)
        img = img.resize((320, 320))
        img = img.rotate(rotation * -90)
        data = img.getdata()
        result = []
        pixelDataIndex = 0
        for i in range(800):
            for p in range(0, 512, 4):
                result.append(data[pixelDataIndex][0])
                result.append(data[pixelDataIndex][1])
                result.append(data[pixelDataIndex][2])
                result.append(0)
                pixelDataIndex += 1
        return result

    def _prepare_gif_file(self, path, rotation):
        """
        path is the path of the gif file
        Rotation is expected as 0 = no rotation, 1 = 90 degrees, 2 = 180 degrees, 3 = 270 degrees
        Gifs are resized to 320x320 and rotated to match the desired orientation
        result is a bytesIo stream
        """
        img = Image.open(path)

        frames = ImageSequence.Iterator(img)

        def prepare_frames(frames):
            for frame in frames:
                resized = frame.copy()
                resized = resized.resize((320, 320))
                resized = resized.rotate(rotation * -90)
                yield resized

        frames = prepare_frames(frames)

        result_img = next(frames)  # Handle first frame separately
        result_img.info = img.info  # Copy sequence info

        result_bytes = io.BytesIO()
        result_img.save(
            result_bytes, format="GIF", save_all=True, append_images=list(frames), loop=0
        )

        return result_bytes.getvalue()

    def _send_data(self, data, bulkInfo):
        """
        sends image or gif to device
        data is an array of bytes to write
        bulk info contains info about the transfer
        """

        assert self.bulk_device, "Cannot find bulk out device"

        self._write_then_read([0x36, 0x03])  # unknown

        buckets = self._query_buckets()  # query all buckets and store their response

        self._write_then_read([0x20, 0x03])  # unknown
        self._write_then_read([0x74, 0x01])  # keepalive
        self._write_then_read([0x70, 0x01])  # unknown
        self._write_then_read([0x74, 0x01])  # keepalive

        bucketIndex = self._find_next_unoccupied_bucket(
            buckets
        )  # find the first unoccupied bucket in the list
        bucketIndex = self._prepare_bucket(
            bucketIndex if bucketIndex != -1 else 0, bucketIndex == -1
        )  # prepare bucket or find a more suitable one

        packetCount = math.floor((math.ceil(len(data) / _BULK_WRITE_LENGTH) + 1) / 2)
        packetCountBytes = list(
            packetCount.to_bytes(2, "little")  # calculates the number of needed packets
        )

        bucketMemoryStart = self._get_bucket_memory_offset(
            buckets, bucketIndex, packetCount
        )  # extracts the bucket starting address

        if bucketMemoryStart == -1:  # cant find a good memory start
            self._delete_all_buckets()
            bucketIndex = 0  # start from byte 0
            bucketMemoryStart = [0x0, 0x0]

        # setup bucket for transfer
        if not self._setup_bucket(
            bucketIndex, bucketIndex + 1, bucketMemoryStart, packetCountBytes
        ):
            _LOGGER.error("Failed to setup bucket for data transfer")

        self._write_then_read([0x36, 0x01, bucketIndex])  # start data transfer

        self._bulk_write(
            [
                0x12,
                0xFA,
                0x01,
                0xE8,
                0xAB,
                0xCD,
                0xEF,
                0x98,
                0x76,
                0x54,
                0x32,
                0x10,
            ]  # first bulk write message contains a standard part and information about the transfer
            + bulkInfo
        )

        for i in range(0, len(data), _BULK_WRITE_LENGTH):  # start sending data in 512mb chunks
            self._bulk_write(list(data[i : i + _BULK_WRITE_LENGTH]))

        self._write([0x36, 0x02])  # end data transfer
        if not self._switch_bucket(bucketIndex):  # switch to newly written bucket
            _LOGGER.error("Failed to switch active bucket")

    def _query_buckets(self):
        """
        Queries all 16 buckets and stores their response
        Response in structures as follow:
        - standard part (14 bytes) - unknown
        ---- following is all 0x0 if bucket is unoccupied
        - bucket index (1 byte)
        - asset index (1 byte) - same as bucket index + 1
        - 0x2 (1 byte) - unknown
        - starting memory address (2 bytes) - address sometimes changes so must be read from here
        - memory size (2 bytes) - size sometimes changes so must be read from here
        - 0x1 (1 byte) - unknown
        - 0x0|0x1 (1 byte) - most likely used/unused but could also be something else
        """
        buckets = {}
        for bI in range(16):
            response = self._write_then_read([0x30, 0x04, bI])  # query bucket
            buckets[bI] = response
        return buckets

    def _find_next_unoccupied_bucket(self, buckets):
        """
        finds the first available unoccupied bucket
        buckets are unoccupied when bytes 14 onward are 0x0
        returns -1 if unoccupied buckets are found
        """
        for bucketIndex, bucketInfo in buckets.items():
            if not any(bucketInfo[15:]):
                return bucketIndex
        return -1

    def _get_bucket_memory_offset(self, buckets, bucketIndex, packetCount):
        """
        returns the memory start address for the selected bucket
        memory offset is calculated by first checking if the bucket can already
        accommodate the new data, this avoids any additional calculations if uploading the same or smaller image
        otherwise, we check if we can expand the current bucket without overlapping the memory space of any other bucket
        otherwise, we set the offset after max utilized memory if there is space left on the device
        otherwise, we check if there is space at the beginning of the memory space
        finally, if all else fails then we clear the device and start over
        """

        currentBucket = buckets[bucketIndex]
        currentBucketOffset = int.from_bytes([currentBucket[17], currentBucket[18]], "little")
        currentBucketSize = int.from_bytes([currentBucket[19], currentBucket[20]], "little")

        # check if we can fit content in existing bucket space
        if packetCount <= currentBucketSize:
            return [currentBucket[17], currentBucket[18]]

        # find max byte number
        minOccupiedByte = currentBucketOffset
        maxOccupiedByte = 0
        existingBucketWithinRange = False
        for bI in buckets:
            bucket = buckets[bI]
            startByte = int.from_bytes([bucket[17], bucket[18]], "little")
            endByte = startByte + int.from_bytes([bucket[19], bucket[20]], "little")
            if endByte > maxOccupiedByte:
                maxOccupiedByte = endByte
            if startByte < minOccupiedByte:
                minOccupiedByte = startByte
            if (
                (startByte > currentBucketOffset and startByte < currentBucketOffset + packetCount)
                or (startByte < currentBucketOffset and endByte > startByte)
                or (startByte == currentBucketOffset and bI != bucketIndex)
            ):
                existingBucketWithinRange = True

        # check if we can use current offset without overlapping other buckets
        if not existingBucketWithinRange:
            return [currentBucket[17], currentBucket[18]]

        # check if we would exceed available memory if we put data at the end
        if maxOccupiedByte + packetCount < _LCD_TOTAL_MEMORY:
            return list(maxOccupiedByte.to_bytes(2, "little"))

        # if the lowest used byte is more than zero and we can fit the data then start from zero
        if packetCount < minOccupiedByte:
            return [0x0, 0x0]

        # if all else fails return -1 to reset and start over
        return -1

    def _prepare_bucket(self, bucketIndex, bucketFilled):
        """
        if a bucket delete returns 0x9 then try next bucket
        if bucket already had data then delete it twice
        """
        assert bucketIndex < 16, "reached max bucket"
        delete_response = self._delete_bucket(bucketIndex)
        if not delete_response:
            return self._prepare_bucket(bucketIndex + 1, True)
        else:
            if bucketFilled:
                return self._prepare_bucket(bucketIndex, False)
        return bucketIndex

    def _delete_bucket(self, bucketIndex):
        """
        deletes bucket, returns true if successful, false otherwise
        """
        self._write([0x32, 0x2, bucketIndex])

        def parse_delete_result(msg):
            return msg[14] == 0x1

        return self._read_until_first_match({b"\x33\x02": parse_delete_result})

    def _delete_all_buckets(self):
        """
        Switches to liquid mode then deletes all buckets
        """
        self._switch_bucket(0, 2)  # switch to liquid mode
        for bI in range(16):
            self._delete_bucket(bI)  # delete bucket

    def _switch_bucket(self, bucketIndex, mode=0x4):
        """
        switches active bucket, returns true if successful, false otherwise
        """
        response = self._write_then_read([0x38, 0x1, mode, bucketIndex])
        return response[14] == 0x1

    def _setup_bucket(self, startBucketIndex, endBucketIndex, startingMemoryAddress, memorySize):
        """
        sets bucket for transmission, returns true if successful, false otherwise
        """
        response = self._write_then_read(
            [
                0x32,
                0x1,
                startBucketIndex,
                endBucketIndex,
                startingMemoryAddress[0],
                startingMemoryAddress[1],
                memorySize[0],
                memorySize[1],
                0x1,
            ]
        )
        return response[14] == 0x1



"""Test backward compatibility with liquidctl 1.1.x."""

import pytest

import usb

from liquidctl.driver.kraken2 import Kraken2
from liquidctl.driver.usb import hid, HidapiDevice


class _MockPyUsbHandle(usb.core.Device):
    def __init__(self, serial_number):
        self.idVendor = 0x1e71
        self.idProduct = 0x170e
        self._serial_number = serial_number

        class MockResourceManager():
            def dispose(self, *args, **kwargs):
                pass

        self._ctx = MockResourceManager()


def _mock_enumerate(vendor_id=0, product_id=0):
    return [
        {'vendor_id': vendor_id, 'product_id': product_id, 'serial_number': '_21', 'path': b'/_21'},
        {'vendor_id': vendor_id, 'product_id': product_id, 'serial_number': '_89', 'path': b'/_89'}
    ]


def test_construct_with_raw_pyusb_handle(monkeypatch):
    monkeypatch.setattr(hid, 'enumerate', _mock_enumerate)
    pyusb_handle = _MockPyUsbHandle(serial_number='_89')
    liquidctl_device = Kraken2(pyusb_handle, 'Some device')
    assert liquidctl_device.device.vendor_id == pyusb_handle.idVendor, \
        '<driver instance>.device points to incorrect physical device'
    assert liquidctl_device.device.product_id == pyusb_handle.idProduct, \
        '<driver instance>.device points to incorrect physical device'
    assert liquidctl_device.device.serial_number == pyusb_handle.serial_number, \
        '<driver instance>.device points to different physical unit'
    assert isinstance(liquidctl_device.device, HidapiDevice), \
        '<driver instance>.device not properly converted to HidapiDevice instance'



"""Base bus and driver API.

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

# uses the psf/black style


class BaseDriver:
    """Base driver API.

    All drivers are expected to implement this API for compatibility with the
    liquidctl CLI or other thirdy party tools.

    Drivers will automatically implement the context manager protocol, but this
    should only be used from a call to `connect`.

    Example:

        for dev in <Driver>.find_supported_devices():
            with dev.connect():
                print(dev.get_status())
                if dev.serial_number == '49385027ZP':
                    dev.set_fixed_speed('fan3', 42)

    """

    @classmethod
    def find_supported_devices(cls, **kwargs):
        """Find and bind to compatible devices.

        Returns a list of bound driver instances.
        """

        raise NotImplementedError()

    def connect(self, **kwargs):
        """Connect to the device.

        Returns `self`.

        ## Notes for driver authors

        Procedure before any read or write operation can be performed.
        Typically a handshake between driver and device.
        """

        raise NotImplementedError()

    def initialize(self, **kwargs):
        """Initialize the device and the driver.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.

        ## Notes for driver authors

        Apart from `connect()`, some devices might require a one-time
        initialization procedure after powering on, or to detect hardware
        changes.  This should be called *after* connecting to the device.
        """

        raise NotImplementedError()

    def disconnect(self, **kwargs):
        """Disconnect from the device.

        ## Notes for driver authors

        Procedure before the driver can safely unbind from the device.
        Typically just cleanup.
        """

        raise NotImplementedError()

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        raise NotImplementedError()

    def set_color(self, channel, mode, colors, **kwargs):
        """Set the color mode for a specific channel."""

        raise NotImplementedError()

    def set_screen(self, channel, mode, value, **kwargs):
        """Set the screen mode and content.

        Unstable.
        """

        raise NotImplementedError()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Set channel to follow a speed duty profile."""

        raise NotImplementedError()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""

        raise NotImplementedError()

    @property
    def description(self):
        """Human readable description of the corresponding device."""

        raise NotImplementedError()

    @property
    def vendor_id(self):
        """Numeric vendor identifier, or None if N/A."""

        raise NotImplementedError()

    @property
    def product_id(self):
        """Numeric product identifier, or None if N/A."""

        raise NotImplementedError()

    @property
    def release_number(self):
        """Device versioning number, or None if N/A.

        In USB devices this is bcdDevice.
        """

        raise NotImplementedError()

    @property
    def serial_number(self):
        """Serial number reported by the device, or None if N/A."""

        raise NotImplementedError()

    @property
    def bus(self):
        """Bus the device is connected to, or None if N/A."""

        raise NotImplementedError()

    @property
    def address(self):
        """Address of the device on the corresponding bus, or None if N/A.

        This typically depends on the bus enumeration order.
        """

        raise NotImplementedError()

    @property
    def port(self):
        """Physical location of the device, or None if N/A.

        This typically refers to a USB port, which is *not* dependent on bus
        enumeration order.  However, a USB port is hub-specific, and hubs can
        be chained.  Thus, for USB devices, this returns a tuple of port
        numbers, from the root hub to the parent of the connected device.
        """

        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.disconnect()


class BaseBus:
    """Base bus API."""

    def find_devices(self, **kwargs):
        """Find compatible devices and yield corresponding driver instances."""
        return


def find_all_subclasses(cls):
    """Recursively find loaded subclasses of `cls`.

    Returns a set of subclasses of `cls`.
    """
    sub = set(cls.__subclasses__())
    return sub.union([s for c in cls.__subclasses__() for s in find_all_subclasses(c)])
